import warnings

from finetune.util.logging import truncate_text


def assign_associations(labels, associations, none_value):
    idx_lookups = [{} for _ in labels]
    for i, (doc_label, doc_association) in enumerate(zip(labels, associations)):
        active_label_idx = -1
        for label, association in zip(doc_label, doc_association):
            if label == none_value:
                continue
            active_label_idx += 1
            for bpe_idx, _, _, _ in association:
                idx_lookups[i][bpe_idx] = active_label_idx

    all_candidates = []

    for idx_lookup, doc_label, doc_association in zip(idx_lookups, labels, associations):
        candidates = {}
        if doc_label == none_value:
            continue

        for association in doc_association:
            for bpe_idx, candidate_idx, candidate_label, candidate_prob in association:
                if candidate_label == none_value or candidate_idx not in idx_lookup:
                    continue

                if idx_lookup[bpe_idx] not in candidates:
                    candidates[idx_lookup[bpe_idx]] = []

                candidates[idx_lookup[bpe_idx]].append((idx_lookup[candidate_idx], candidate_label, candidate_prob))

        # TODO some how sample these candidates eg maximum probabilities, to fit some schema
        candidates = {k: max(v, key=lambda x: x[2]) for k, v in candidates.items()} # for now just pick maximum prob
        all_candidates.append(candidates)
    return all_candidates


def finetune_to_indico_sequence(raw_texts, subseqs, labels, encoder=None, probs=None, none_value=None,
                                subtoken_predictions=False, associations=None):
    """
    Maps from the labeled substring format into the 'indico' format. This is the exact inverse operation to
    :meth indico_to_finetune_sequence:.

    The indico format is as follows:
        Raw text for X,
        Labels as a list of dicts, with each dict in the form:
        {
            'start': <Character index of the start of the labeled sequence>,
            'end': <Character index of the end of the labeled sequence>,
            'label': <A categorical label (int or string) that represents the category of the subsequence,
            'text': <Optionally, a field with the subsequence contained between the start and end.
        }

    The Labeled substring, or finetune internal, format is as follows.
    Each item of the data is a list strings of the form:
        ["The quick brown", "fox", "jumped over the lazy", ...]
    With the corresponding labels:
        ["PAD", "animal", "PAD", ...]

    It is the :param none_value: that is used to populate the PAD labels.
    :param data: A list of segmented text of the form list(list(str))
    :param labels: Categorical labels for each sub-string in data.
    :param none_value: The none value used to encode the input format.
    :return: Texts, annoatations both in the 'indico' format.
    """
    annotations = []
    if associations is not None:
        assoc_cleaned = assign_associations(labels, associations, none_value)
    else:
        assoc_cleaned = [None] * len(raw_texts)

    encoded_docs = encoder._encode(raw_texts)
    loop_vals = zip(raw_texts, subseqs, labels, probs or [None] * len(raw_texts), assoc_cleaned)
    for doc_idx, (raw_text, doc_seq, label_seq, prob_seq, associations_seq) in enumerate(loop_vals):
        tokens = encoded_docs.tokens[doc_idx]
        token_ends = encoded_docs.char_locs[doc_idx]
        token_lengths = [encoder._token_length(token) for token in tokens]
        token_starts = [end - length for end, length in zip(token_ends, token_lengths)]
        n_tokens = len(tokens)

        doc_annotations = []
        annotation_ranges = set()
        start_idx = 0
        end_idx = 0
        raw_annotation_start = 0
        for i, (sub_str, raw_label, confidences) in enumerate(zip(doc_seq, label_seq, prob_seq or [None] * len(doc_seq))):
            if not isinstance(raw_label, tuple):
                multi_label = False
                label_list = [raw_label]
            else:
                multi_label = True
                label_list = raw_label

            for label in label_list:
                stripped_text = sub_str.strip()

                raw_annotation_start = raw_text.find(stripped_text, raw_annotation_start)
                raw_annotation_end = raw_annotation_start + len(stripped_text)
                for i, item in enumerate(doc_annotations):
                    if item["label"] == label and raw_annotation_start - item["end"] <= 1:
                        raw_annotation_start = item["start"]
                        doc_annotations.pop(i)
                        break

                if raw_annotation_start == -1:
                    warnings.warn("Failed to find predicted sequence in text: {}.".format(
                        truncate_text(stripped_text)
                    ))
                    continue

                annotation_start = raw_annotation_start
                annotation_end = raw_annotation_end

                # if we don't want to allow subtoken predictions, adjust start and end to match
                # the start and ends of the nearest full tokens
                if not subtoken_predictions:
                    if multi_label:
                        start_idx = 0
                        end_idx = 0
                    if label != none_value:
                        # round to nearest token
                        while start_idx < n_tokens and annotation_start >= token_starts[start_idx]:
                            start_idx += 1
                        annotation_start = token_starts[start_idx - 1]
                        while end_idx < (n_tokens - 1) and annotation_end > token_ends[end_idx]:
                            end_idx += 1
                        annotation_end = token_ends[end_idx]

                text = raw_text[annotation_start:annotation_end]
                if label != none_value:
                    annotation = {
                        "start": int(annotation_start),
                        "end": int(annotation_end),
                        "label": label,
                        "text": text
                    }
                    if associations_seq is not None and len(doc_annotations) in associations_seq:
                        index, relationship, prob = associations_seq[len(doc_annotations)]
                        annotation["associations"] = {
                            "index": index,
                            "relationship": relationship,
                            "prob": prob
                        }
                    if confidences is not None:
                        annotation["confidence"] = confidences

                    # prevent duplicate annotation edge case
                    if (annotation_start, annotation_end, label) not in annotation_ranges:
                        annotation_ranges.add((annotation_start, annotation_end, label))
                        doc_annotations.append(annotation)

        doc_annotations = sorted([dict(items) for items in doc_annotations], key=lambda x: x['start'])
        annotations.append(doc_annotations)
    return raw_texts, annotations


def indico_to_finetune_sequence(texts, labels=None, encoder=None, multi_label=True, none_value=None,
                                subtoken_labels=False):
    """
    Maps from the 'indico' format sequence labeling data. Into a labeled substring format. This is the exact inverse of
    :meth finetune_to_indico_sequence:.

    The indico format is as follows:
        Raw text for X,
        Labels as a list of dicts, with each dict in the form:
         labeled sequence>,
            'end': <Character index of the end of the labeled sequence>,
            'label': <A categorical label (int or string) that represents the category of the subsequence,
            'text': <A field containing the sub-sequence contained between the start and end.
        }

    The Labeled substring, or finetune internal, format is as follows.
    Each item of the data is a list strings of the form:{
            'start': <Character index of the start of the
        ["The quick brown", "fox", "jumped over the lazy", ...]
    With the corresponding labels:
        ["PAD", "animal", "PAD", ...]

    It is the :param none_value: that is used to populate the PAD labels.

    :param texts: A list of raw text.
    :param labels: A list of targets of the form list(list(dict))).
    :param none_value: A categorical label to use as the none value.
    :return: Segmented Text, Labels of the form described above.
    """
    all_subseqs = []
    all_labels = []
    all_association_idx = []
    all_association_type = []
    all_idxs = []

    # placeholder for inference time
    if labels is None:
        labels = [[]] * len(texts)

    encoded_docs = encoder._encode(texts)

    for doc_idx, (text, label_seq) in enumerate(zip(texts, labels)):
        tokens = encoded_docs.tokens[doc_idx]
        token_ends = encoded_docs.char_locs[doc_idx]
        token_lengths = [encoder._token_length(token) for token in tokens]
        token_starts = [end - length for end, length in zip(token_ends, token_lengths)]
        n_tokens = len(tokens)

        label_seq = sorted(label_seq, key=lambda x: x["start"])
        last_loc = 0
        doc_subseqs = []
        doc_labels = []
        doc_association_idx = []
        doc_association_type = []
        doc_current_label_idx = []

        for i, annotation in enumerate(label_seq):
            start = annotation["start"]
            end = annotation["end"]
            label = annotation["label"]
            annotation_text = annotation.get("text")
            if "association" in annotation:
                association_idx = annotation["association"]["index"]
                association_type = annotation["association"]["relationship"]
            else:
                association_idx = -1
                association_type = none_value

            if annotation_text is not None and text[start:end] != annotation_text:
                raise ValueError(
                    "Annotation text does not match text specified by `start` and `end` indexes. "
                    "Text provided: `{}`.  Text extracted: `{}`.".format(
                        annotation_text,
                        text[start:end]
                    )
                )

            if not subtoken_labels:
                if label != none_value:
                    # round to nearest token
                    while start > 0 and start not in token_starts:
                        start -= 1
                    while end < len(text) and end not in token_ends:
                        end += 1

            if start > last_loc:
                doc_subseqs.append(text[last_loc:start])
                if multi_label:
                    doc_labels.append([none_value])
                else:
                    doc_labels.append(none_value)
                    doc_association_idx.append(-1)
                    doc_association_type.append(none_value)
                    doc_current_label_idx.append(-2)

            j = len(doc_labels) - 1
            split_dist = last_loc - end
            skip_end = 0
            if split_dist > 0:
                j -= 1
                if len(doc_subseqs[-1]) != split_dist:
                    dual_label_sub_seq = doc_subseqs[-1][-split_dist:]
                    doc_subseqs[-1] = doc_subseqs[-1][:-split_dist]
                    doc_subseqs.append(dual_label_sub_seq)
                    doc_labels.append(doc_labels[-1][:])
                    j -= 1

                skip_end = len(doc_subseqs[-1])

            if start < last_loc - skip_end:
                if not multi_label:
                    raise ValueError("Overlapping annotations requires the multi-label model")
                else:
                    split_dist = last_loc - start - skip_end
                    while split_dist >= len(doc_subseqs[j]):
                        doc_labels[j].append(label)
                        split_dist -= len(doc_subseqs[j])
                        j -= 1

                    if split_dist > 0:
                        dual_label_sub_seq = doc_subseqs[j][-split_dist:]
                        doc_subseqs[j] = doc_subseqs[j][:-split_dist]
                        doc_subseqs.insert(j + 1, dual_label_sub_seq)
                        doc_labels.insert(j + 1, doc_labels[j][:] + [label])

                    start = last_loc
            if start >= end:
                # degenerate label
                last_loc = max(start, end)
                continue

            doc_subseqs.append(text[start:end])
            if multi_label:
                doc_labels.append([label])
            else:
                doc_labels.append(label)
                doc_association_idx.append(association_idx)
                doc_association_type.append(association_type)
                doc_current_label_idx.append(i)

            last_loc = end

        if last_loc != len(text):
            doc_subseqs.append(text[last_loc:])
            if multi_label:
                doc_labels.append([none_value])
            else:
                doc_labels.append(none_value)
                doc_association_idx.append(-1)
                doc_association_type.append(none_value)
                doc_current_label_idx.append(-2)

        all_subseqs.append(doc_subseqs)
        all_labels.append(doc_labels)
        all_association_idx.append(doc_association_idx)
        all_association_type.append(doc_association_type)
        all_idxs.append(doc_current_label_idx)

    return all_subseqs, all_labels, all_association_type, all_association_idx, all_idxs



import functools

import tensorflow as tf
from tensorflow.python.framework import function, ops


def fn_with_custom_grad(grad_fn, use_global_vars=False, use_entire_scope=False):
    """Decorator to create a subgraph with a custom gradient function.
    The subgraph created by the decorated function is NOT put in a Defun and so
    does not suffer from the limitations of the Defun (all subgraph ops on the
    same device, no summaries).
    Args:
      grad_fn: function with signature
        (inputs, variables, outputs, output_grads) -> (grad_inputs, grad_vars),
        all of which are lists of Tensors.
      use_global_vars: if True, variables will be the global variables created.
        If False, will be the trainable variables.
    Returns:
      Decorator for function such that the gradient is defined by grad_fn.
    """

    def dec(fn):
        @functools.wraps(fn)
        def wrapped(*args):
            return _fn_with_custom_grad(
                fn, args, grad_fn, use_global_vars=use_global_vars, use_entire_scope=use_entire_scope)

        return wrapped

    return dec


def _fn_with_custom_grad(fn, inputs, grad_fn, use_global_vars=False, use_entire_scope=False):
    """Create a subgraph with a custom gradient.
    Args:
      fn: function that takes inputs as arguments and produces 1 or more Tensors.
      inputs: list<Tensor>, will be passed as fn(*inputs).
      grad_fn: function with signature
        (inputs, vars, outputs, output_grads) -> (grad_inputs, grad_vars),
        all of which are lists of Tensors.
      use_global_vars: if True, variables will be the global variables created.
        If False, will be the trainable variables.
    Returns:
      fn(*inputs)
    """
    vs = tf.get_variable_scope()

    # Use a function attribute to store the local variables. This means that on graph rebuild,
    # When variables already exist we know what variables relate to this function.
    get_vars_fn = (vs.global_variables if use_global_vars else vs.trainable_variables)
    len_before_vars = len(get_vars_fn())
    inputs = list(inputs)
    outputs = fn(*inputs)
    if use_entire_scope:
        train_vars = get_vars_fn()
    else:
        train_vars = get_vars_fn()[len_before_vars:]

    if grad_fn is None:
        return outputs

    if not isinstance(outputs, (tuple, list)):
        outputs = [outputs]
    outputs = list(outputs)

    defun_inputs = [inputs, train_vars, outputs]

    def custom_grad_fn(op, *dys):
        """Custom grad fn applying grad_fn for identity Defun."""
        fn_inputs, fn_vars, fn_outputs = tf.contrib.framework.nest.pack_sequence_as(
            defun_inputs, list(op.inputs))
        dys = list(dys)
        assert len(fn_outputs) == len(outputs)
        assert len(fn_outputs) == len(dys)

        grad_inputs, grad_vars = grad_fn(fn_inputs, fn_vars, fn_outputs, dys)
        grad_outputs = [None] * len(fn_outputs)
        return tuple(grad_inputs + grad_vars + grad_outputs)

    # The Defun takes as input the original inputs, the trainable variables
    # created in fn, and the outputs. In the forward it passes through the
    # outputs. In the backwards, it produces gradients for the original inputs
    # and the trainable variables.
    in_types = [t.dtype for t in inputs]
    out_types = [t.dtype for t in outputs]
    var_types = [t.dtype for t in train_vars]

    @function.Defun(
        *(in_types + var_types + out_types),
        func_name="identity_custom_grad%d" % ops.uid(),
        python_grad_func=custom_grad_fn,
        shape_func=lambda _: [t.get_shape() for t in outputs])
    def identity(*args):
        _, _, outs = tf.contrib.framework.nest.pack_sequence_as(defun_inputs, args)
        return tuple([tf.identity(t) for t in outs])

    flat_inputs = tf.contrib.framework.nest.flatten(defun_inputs)
    id_out = identity(*flat_inputs)

    return id_out


def recompute_grad(fn, use_entire_scope):
    """Decorator that recomputes the function on the backwards pass.
    Args:
      fn: a function that takes Tensors (all as positional arguments) and returns
        a tuple of Tensors.
    Returns:
      A wrapped fn that is identical to fn when called, but its activations will
      be discarded and recomputed on the backwards pass (i.e. on a call to
      tf.gradients).
      :param use_entire_scope:
    """

    @functools.wraps(fn)
    def wrapped(*args):
        return _recompute_grad(fn, args, use_entire_scope=use_entire_scope)

    return wrapped


def underlying_variable_ref(t):
    """Find the underlying variable ref.
    Traverses through Identity, ReadVariableOp, and Enter ops.
    Stops when op type has Variable or VarHandle in name.
    Args:
      t: a Tensor
    Returns:
      a Tensor that is a variable ref, or None on error.
    """
    while t.op.type in ["Identity", "ReadVariableOp", "Enter"]:
        t = t.op.inputs[0]

    op_type = t.op.type
    if "Variable" in op_type or "VarHandle" in op_type:
        return t
    else:
        return None


def _recompute_grad(fn, args, use_entire_scope):
    """See recompute_grad."""

    cached_vs = []
    cached_arg_scope = []

    def grad_fn(inputs, variables, outputs, output_grads):
        """Recompute outputs for gradient computation."""
        del outputs
        variables = [underlying_variable_ref(v) for v in variables]
        # Recompute outputs
        with tf.control_dependencies(output_grads):
            with tf.contrib.framework.arg_scope(cached_arg_scope[0]):
                with tf.variable_scope(cached_vs[0], reuse=True):
                    outputs = fn(*inputs)

        if not isinstance(outputs, (list, tuple)):
            outputs = [outputs]
        outputs = list(outputs)
        grads = tf.gradients(outputs, inputs + variables, output_grads)
        grad_inputs = grads[:len(inputs)]
        grad_vars = grads[len(inputs):]
        return grad_inputs, grad_vars

    @fn_with_custom_grad(grad_fn, use_entire_scope=use_entire_scope)
    def fn_with_recompute(*args):
        cached_vs.append(tf.get_variable_scope())
        cached_arg_scope.append(tf.contrib.framework.current_arg_scope())
        return fn(*args)

    return fn_with_recompute(*args)


import os

from finetune.base_models import SourceModel
from finetune.base_models.bert.encoder import BERTEncoder, BERTEncoderMultuilingal, BERTEncoderLarge
from finetune.base_models.bert.featurizer import bert_featurizer


class BERTModelCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoder
    featurizer = bert_featurizer
    settings = {
        'n_embed': 768,
        'n_epochs': 8,
        'n_heads': 12,
        'n_layer': 12,
        'act_fn': "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        'bert_intermediate_size': 3072,
        "base_model_path": os.path.join("bert", "bert_small_cased.jl"),
    }


class BERTModelLargeCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoderLarge
    featurizer = bert_featurizer
    settings = {
        'n_embed': 1024,
        'n_epochs': 8,
        'n_heads': 16,
        'n_layer': 24,
        'num_layers_trained': 24,
        'act_fn': "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        'bert_intermediate_size': 4096,
        "base_model_path": os.path.join("bert", "bert_large_cased.jl"),
    }


class BERTModelMultilingualCased(SourceModel):
    is_bidirectional = True
    encoder = BERTEncoderMultuilingal
    featurizer = bert_featurizer
    settings = {
        'n_embed': 768,
        'n_epochs': 8,
        'n_heads': 12,
        'n_layer': 12,
        'act_fn': "gelu",
        "lr_warmup": 0.1,
        "lr": 1e-5,
        "l2_reg": 0.01,
        'bert_intermediate_size': 3072,
        "base_model_path": os.path.join("bert", "bert_small_multi_cased.jl"),
    }



import os
import logging
from pathlib import Path

import numpy as np

from sklearn.model_selection import train_test_split

from finetune import Classifier
from finetune.datasets import Dataset, generic_download

logging.basicConfig(level=logging.DEBUG)

SST_FILENAME = "SST-binary.csv"
DATA_PATH = os.path.join('Data', 'Classify', SST_FILENAME)
CHECKSUM = "02136b7176f44ff8bec6db2665fc769a"


class StanfordSentimentTreebank(Dataset):

    def __init__(self, filename=None, **kwargs):
        super().__init__(filename=(filename or DATA_PATH), **kwargs)

    def md5(self):
        return CHECKSUM

    def download(self):
        """
        Download Stanford Sentiment Treebank to data directory
        """
        path = Path(self.filename)
        path.parent.mkdir(parents=True, exist_ok=True)
        generic_download(
            url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
            text_column="Text",
            target_column="Target",
            filename=SST_FILENAME
        )


if __name__ == "__main__":
    # Train and evaluate on SST
    dataset = StanfordSentimentTreebank(nrows=200).dataframe
    pre_train_generator = lambda: iter(StanfordSentimentTreebank(nrows=5000).dataframe.Text.values)
    model = Classifier(n_epochs=3, batch_size=2, lr_warmup=0.1, tensorboard_folder='.tensorboard')
    trainX, testX, trainY, testY = train_test_split(dataset.Text.values, dataset.Target.values, test_size=0.3, random_state=42)
    model.config.dataset_size = 5000
    model.config.val_size = 100
    model.config.val_interval = 1000
    model.config.batch_size = 5
    model.fit(pre_train_generator)
    model.config.val_size = None
    model.config.val_interval = None
    #model.config.dataset_size = 1000 # This is automatically set as trainX has len
    model.config.batch_size = 2
    model.fit(trainX, trainY)
    accuracy = np.mean(model.predict(testX) == testY)
    print('Test Accuracy: {:0.2f}'.format(accuracy))



import logging
import functools

import numpy as np
import tensorflow as tf
from tensorflow.contrib.opt.python.training.weight_decay_optimizers import AdamWOptimizer

from finetune.nn.target_blocks import language_model
from finetune.util.text_generation import sample_with_temperature
from finetune.optimizers.zero_grad import dont_optimize_zeros
from finetune.optimizers.gradient_accumulation import get_grad_accumulation_optimizer
from finetune.optimizers.learning_rate_schedules import schedules
from finetune.optimizers.adamax import AdamaxWOptimizer
from finetune.util.imbalance import class_weight_tensor
from finetune.errors import FinetuneError
from finetune.base_models import GPTModel, GPTModelSmall

LOGGER = logging.getLogger('finetune')

OPTIMIZERS = {
    "AdamW": AdamWOptimizer,
    "AdamaxW": AdamaxWOptimizer
}

class PredictMode:
    FEATURIZE = "FEAT"
    NORMAL = "NORM"
    PROBAS = "PROBA"
    GENERATE_TEXT = "GEN_TEXT"
    LM_PERPLEXITY = "PERPLEXITY"
    ATTENTION = "ATTENTION"
    SEQUENCE = "SEQUENCE"
    SEQUENCE_PROBAS = "SEQUENCE_PROBA"
    ASSOCIATION = "ASSOCIATION"
    ASSOCIATION_PROBAS = "ASSOCIATION_PROBA"


def get_model_fn(target_model_fn, predict_op, predict_proba_op, build_target_model, build_lm, encoder, target_dim,
                 label_encoder, saver):
    def language_model_op(X, M, params, featurizer_state):
        language_model_state = language_model(
            X=X,
            M=M,
            config=params,
            embed_weights=featurizer_state['embed_weights'],
            hidden=featurizer_state['sequence_features'],
        )

        lm_logits = language_model_state["logits"]

        lm_logit_mask = np.zeros([1, lm_logits.get_shape().as_list()[-1]], dtype=np.float32)
        lm_logit_mask[:, encoder.vocab_size:] = -np.inf

        if "use_extra_toks" in params and not params.use_extra_toks:
            lm_logit_mask[:, encoder.start] = -np.inf
            lm_logit_mask[:, encoder.delimiter] = -np.inf
            lm_logit_mask[:, encoder.clf_token] = -np.inf

        lm_logits += lm_logit_mask
        lm_predict_op = sample_with_temperature(lm_logits, params.lm_temp)
        return lm_predict_op, language_model_state

    def target_model_op(featurizer_state, Y, params, mode, **kwargs):
        weighted_tensor = None
        if params.class_weights is not None:
            weighted_tensor = class_weight_tensor(
                class_weights=params.class_weights,
                target_dim=target_dim,
                label_encoder=label_encoder
            )
        with tf.variable_scope('model/target'):
            target_model_state = target_model_fn(
                config=params,
                featurizer_state=featurizer_state,
                targets=Y,
                n_outputs=target_dim,
                train=(mode == tf.estimator.ModeKeys.TRAIN),
                max_length=params.max_length,
                class_weights=weighted_tensor,
                **kwargs
            )
        return target_model_state

    def _model_fn(features, labels, mode, params):
        if params.base_model.is_bidirectional and build_lm:
            raise ValueError("Bert models do not support functions that require the language model.")
        if not build_target_model:
            lm_loss_coef = 1.
        else:
            lm_loss_coef = params.lm_loss_coef

        estimator_mode = mode
        train = estimator_mode == tf.estimator.ModeKeys.TRAIN
        X = features["tokens"]
        M = features["mask"]
        task_id = features.get("task_id", None)
        Y = labels
        pred_op = None

        with tf.variable_scope(tf.get_variable_scope()):
            train_loss = 0.0

            featurizer_state = params.base_model.get_featurizer(
                X,
                encoder=encoder,
                config=params,
                train=train
            )
            predictions = {PredictMode.FEATURIZE: featurizer_state["features"]}
            if params.base_model in [GPTModel, GPTModelSmall]:
                predictions[PredictMode.ATTENTION] = featurizer_state["attention_weights"]

            if params.base_model in [GPTModel, GPTModelSmall]:
                predictions[PredictMode.ATTENTION] = featurizer_state["attention_weights"]

            if build_target_model:
                target_model_state = target_model_op(
                    featurizer_state=featurizer_state,
                    Y=Y,
                    params=params,
                    mode=mode,
                    task_id=task_id
                )
                if (mode == tf.estimator.ModeKeys.TRAIN or mode == tf.estimator.ModeKeys.EVAL) and Y is not None:
                    target_loss = tf.reduce_mean(target_model_state["losses"])
                    train_loss += (1 - lm_loss_coef) * target_loss
                    tf.summary.scalar("TargetModelLoss", target_loss)
                if mode == tf.estimator.ModeKeys.PREDICT or tf.estimator.ModeKeys.EVAL:
                    logits = target_model_state["logits"]
                    predict_params = target_model_state.get("predict_params", {})
                    if "_threshold" in params:
                        predict_params["threshold"] = params._threshold
                    pred_op = predict_op(logits, **predict_params)

                    if type(pred_op) == tuple:
                        pred_op, pred_proba_op = pred_op
                    else:
                        pred_proba_op = predict_proba_op(logits, **predict_params)

                    if type(pred_op) == dict:
                        predictions.update(pred_op)
                        predictions.update(pred_proba_op)
                    else:
                        predictions[PredictMode.NORMAL] = pred_op
                        predictions[PredictMode.PROBAS] = pred_proba_op

            if build_lm:
                lm_predict_op, language_model_state = language_model_op(X=X, M=M, params=params,
                                                                        featurizer_state=featurizer_state)
                if mode == tf.estimator.ModeKeys.TRAIN or mode == tf.estimator.ModeKeys.EVAL:
                    lm_loss = tf.reduce_mean(language_model_state["losses"])
                    train_loss += lm_loss_coef * lm_loss
                    tf.summary.scalar("LanguageModelLoss", lm_loss)
                if mode == tf.estimator.ModeKeys.PREDICT:
                    predictions[PredictMode.GENERATE_TEXT] = lm_predict_op
                    predictions[PredictMode.LM_PERPLEXITY] = language_model_state["perplexity"]

        if mode == tf.estimator.ModeKeys.TRAIN:
            total_num_steps = params.n_epochs * params.dataset_size//params.batch_size
            lr_decay = lambda lr, global_step: tf.maximum(
                0.0,
                lr * schedules[params.lr_schedule](tf.to_float(global_step) / total_num_steps)
            )

            def optimizer(lr):

                Optimizer = OPTIMIZERS.get(params.optimizer, None)
                if Optimizer is None:
                    raise FinetuneError(
                        "Optimizer must be in {}, not {}".format(list(OPTIMIZERS.keys()), params.optimizer)
                    )

                if params.accum_steps > 1:
                    Optimizer = get_grad_accumulation_optimizer(Optimizer, params.accum_steps)

                opt = Optimizer(
                    learning_rate=lr,
                    beta1=params.b1,
                    beta2=params.b2,
                    epsilon=params.epsilon,
                    weight_decay=params.l2_reg * lr
                )
                
                decay_var_list = [v for v in tf.global_variables() if len(v.get_shape()) > 1 or params.vector_l2]
                opt.apply_gradients = functools.partial(opt.apply_gradients, decay_var_list=decay_var_list)

                if params.dont_optimize_zero_gradients:
                    opt = dont_optimize_zeros(opt)
                
                return opt

            summaries = tf.contrib.layers.OPTIMIZER_SUMMARIES if params.summarize_grads else None
            train_op = tf.contrib.layers.optimize_loss(
                loss=train_loss,
                global_step=tf.train.get_or_create_global_step(),
                learning_rate=params.lr,
                optimizer=optimizer,
                clip_gradients=float(params.max_grad_norm),
                learning_rate_decay_fn=lr_decay,
                increment_global_step=True,
                summaries=summaries
            )

        if mode == tf.estimator.ModeKeys.PREDICT:
            return tf.estimator.EstimatorSpec(
                mode=mode,
                predictions=predictions
            )

        if mode == tf.estimator.ModeKeys.TRAIN:
            return tf.estimator.EstimatorSpec(mode=mode, loss=train_loss, train_op=train_op)

        assert mode == tf.estimator.ModeKeys.EVAL, "The mode is actually {}".format(mode)
        if params.eval_acc and pred_op is not None:
            LOGGER.info("Adding evaluation metrics, Accuracy")
            labels_dense = tf.argmax(labels, -1)
            metrics = {
                "Accuracy":  tf.metrics.accuracy(pred_op, labels_dense)
            }
        else:
            metrics = None

        return tf.estimator.EstimatorSpec(mode=mode, loss=train_loss, eval_metric_ops=metrics)

    return _model_fn



import logging

import numpy as np

from sklearn.model_selection import train_test_split

from finetune import Classifier
from finetune.datasets.stanford_sentiment_treebank import StanfordSentimentTreebank
from finetune.base_models.bert.model import BERTModelCased
logging.basicConfig(level=logging.DEBUG)

if __name__ == "__main__":
    # Train and evaluate on SST
    dataset = StanfordSentimentTreebank(nrows=1000).dataframe
    model = Classifier(val_size=0.0, max_length=64, base_model=BERTModelCased)
    print(model.config.base_model_path)
    trainX, testX, trainY, testY = train_test_split(dataset.Text.values, dataset.Target.values, test_size=0.3, random_state=42)
    model.fit(trainX, trainY)
    accuracy = np.mean(model.predict(testX) == testY)
    print('Test Accuracy: {:0.2f}'.format(accuracy))



import os
import tempfile
import hashlib
import logging
from pathlib import Path

import numpy as np

from sklearn.model_selection import train_test_split

from finetune import Comparison
from finetune.datasets import Dataset, comparison_download

logging.basicConfig(level=logging.DEBUG)

QUORA_SIMILARITY = "Quora.csv"
DATA_PATH = os.path.join('Data', 'Similarity', QUORA_SIMILARITY)
CHECKSUM = "6f2642eec22c1f0f3fc2fb2dfb9dfc38899d7c58bfc9f4c7e4d1a7e3cea15a29"


class QuoraDuplicate(Dataset):
    def __init__(self, filename=None, **kwargs):
        super().__init__(filename=(filename or DATA_PATH), **kwargs)

    @property
    def md5(self):
        return CHECKSUM

    def download(self):
        """
        Download quora duplicate questions dataset.
        """
        path = Path(self.filename)
        path.parent.mkdir(parents=True, exist_ok=True)
        comparison_download(
            url="https://s3.amazonaws.com/enso-data/Quora.csv",
            text_column1="Text1",
            text_column2="Text2",
            target_column="Target",
            filename=QUORA_SIMILARITY
        )

if __name__ == "__main__":
    # Train and evaluate on SST
    dataset = QuoraDuplicate(nrows=5000).dataframe
    model = Comparison(n_epochs=1)
    trainX1, testX1, trainX2, testX2, trainY, testY = train_test_split(
        dataset.Text1.values, dataset.Text2.values, dataset.Target.values,
        test_size=0.3, random_state=42
    )
    model.fit(list(zip(trainX1, trainX2)), trainY)
    accuracy = np.mean(model.predict(list(zip(testX1, testX2))) == testY)
    class_balance = np.mean(testY)
    print('Test Accuracy: {:0.2f} for a {:0.2f} class balance'.format(accuracy, class_balance))



import types
import tensorflow as tf


def dont_optimize_zeros(optimizer):
    """
    Does not apply the optimiser to variables with 0 gradient. This stops accumulators being diluted in adam and does
     not apply weight decay to these either in the case of weight decay optimisers.

    Based on:
        tf.contrib.opt.MultitaskOptimizerWrapper

    :param optimizer:
    :return:
    """
    
    def _is_all_zeros(grad):
        return tf.equal(tf.count_nonzero(grad), 0)
    
    overridden_methods = ('_apply_dense', '_resource_apply_dense', '_apply_sparse', '_resource_apply_sparse')

    def _get_wrapper(fn, opt):
        
        def wrap(self, grad, *args, **kwargs):
            all_zeros = _is_all_zeros(grad)

            def call_fn():
                with tf.control_dependencies([fn(grad, *args, **kwargs)]):
                    return tf.no_op()
            return tf.cond(all_zeros, tf.no_op, call_fn)
        
        wrapper = types.MethodType(wrap, opt)
        return wrapper

    for name in overridden_methods:
        fn = getattr(optimizer, name)
        wrapper = _get_wrapper(fn, optimizer)
        setattr(optimizer, name, wrapper)

    return optimizer



from abc import ABCMeta, abstractmethod


class SourceModel(metaclass=ABCMeta):
    is_bidirectional = True

    @property
    @abstractmethod
    def encoder(cls):
        raise NotImplementedError

    @property
    @abstractmethod
    def featurizer(cls):
        raise NotImplementedError

    @property
    @abstractmethod
    def settings(cls):
        raise NotImplementedError

    @classmethod
    def get_encoder(cls, **kwargs):
        return cls.encoder(**kwargs)

    @classmethod
    def get_featurizer(cls, X, encoder, config, train=False, reuse=None):
        return cls.featurizer(X, encoder, config, train=train, reuse=reuse)


from finetune.base_models.gpt.model import GPTModel, GPTModelSmall
from finetune.base_models.gpt2.model import GPT2Model
from finetune.base_models.textcnn.model import TextCNNModel
from finetune.base_models.bert.model import BERTModelCased

# aliases
GPT = GPTModel
GPT2 = GPT2Model
TextCNN = TextCNNModel
BERT = BERTModelCased




import tensorflow as tf
from finetune.base_models.gpt.featurizer import dropout, embed


def textcnn_featurizer(X, encoder, config, train=False, reuse=None):
    """
    The transformer element of the finetuning model. Maps from tokens ids to a dense, embedding of the sequence.

    :param X: A tensor of token indexes with shape [batch_size, sequence_length, token_idx]
    :param encoder: A TextEncoder object.
    :param config: A config object, containing all parameters for the featurizer.
    :param train: If this flag is true, dropout and losses are added to the graph.
    :param reuse: Should reuse be set within this scope.
    :return: A dict containing;
        embed_weights: the word embedding matrix.
        features: The output of the featurizer_final state.
        sequence_features: The output of the featurizer at each timestep.
    """
    initial_shape = tf.shape(X)
    X = tf.reshape(X, shape=tf.concat(([-1], initial_shape[-2:]), 0))

    with tf.variable_scope('model/featurizer', reuse=reuse):
        embed_weights = tf.get_variable(
            name="we",
            shape=[encoder.vocab_size + config.max_length, config.n_embed_featurizer],
            initializer=tf.random_normal_initializer(stddev=config.weight_stddev)
        )
        if config.train_embeddings:
            embed_weights = dropout(embed_weights, config.embed_p_drop, train)
        else:
            embed_weights = tf.stop_gradient(embed_weights)

        X = tf.reshape(X, [-1, config.max_length, 2])

        # we remove positional embeddings from the model
        h = embed(X[:, :, :1], embed_weights)

        # keep track of the classify token
        clf_token = encoder['_classify_']
        # Convolutional Layer (this is all the same layer, just different filter sizes)
        pool_layers = []
        conv_layers = []
        for i, kernel_size in enumerate(config.kernel_sizes):
            conv = tf.layers.conv1d(
                inputs=h,
                filters=config.num_filters_per_size,
                kernel_size=kernel_size,
                padding='same',
                activation=tf.nn.relu,
                name='conv' + str(i),
                kernel_initializer=tf.initializers.glorot_normal
            )
            conv_layers.append(conv)
            # mask out the values past the classify token before performing pooling
            pool_idx = tf.cast(tf.argmax(tf.cast(tf.equal(X[:, :, 0], clf_token), tf.float32), 1), tf.int32)
            # mask is past the classify token (i.e. make those results extremely negative)
            mask = tf.expand_dims(1.0 - tf.sequence_mask(pool_idx, maxlen=tf.shape(conv)[1], dtype=tf.float32), -1)
            pool = tf.reduce_max(conv + mask * -1e9, 1)
            pool_layers.append(pool)

        # Concat the output of the convolutional layers for use in sequence embedding
        conv_seq = tf.concat(conv_layers, axis=2)
        seq_feats = tf.reshape(conv_seq, shape=[-1, config.max_length, config.n_embed])
        # Concatenate the univariate vectors as features for classification
        clf_h = tf.concat(pool_layers, axis=1)
        clf_h = tf.reshape(clf_h, shape=tf.concat((initial_shape[: -2], [config.n_embed]), 0))

        # note that, due to convolution and pooling, the dimensionality of the features is much smaller than in the
        # transformer base models
        return {
            'embed_weights': embed_weights,
            'features': clf_h,  # [batch_size, n_embed] for classify, [batch_size, 1, n_embed] for comparison, etc.
            'sequence_features': seq_feats,  # [batch_size, seq_len, n_embed]
            'pool_idx': pool_idx  # [batch_size]
        }



import tensorflow as tf
from tensorflow.python.distribute import device_util
from tensorflow.contrib.distribute import ParameterServerStrategy
from tensorflow.contrib.distribute.python.parameter_server_strategy import ParameterServerExtended
from tensorflow.python.distribute import cross_device_ops as cross_device_ops_lib


_LOCAL_CPU = "/device:CPU:0"
_LOCAL_GPU_0 = "/device:GPU:0"


class PatchedParameterServerExtended(ParameterServerExtended):
    """Implementation of ParameterServerStrategy."""

    def __init__(self, container_strategy, num_gpus_per_worker, visible_gpus):
        super(ParameterServerExtended, self).__init__(container_strategy)
        self._num_gpus_per_worker = num_gpus_per_worker
        self._visible_gpus = visible_gpus
        self._initialize_local(num_gpus_per_worker)

        # We typically don't need to do all-reduce in this strategy.
        self._cross_device_ops = (
            cross_device_ops_lib.ReductionToOneDeviceCrossDeviceOps(
                reduce_to_device=_LOCAL_CPU
            )
        )

    def _initialize_local(self, num_gpus_per_worker):
        """Initialize internal devices for local training."""
        self._worker_device = device_util.canonicalize("/device:CPU:0")
        # Define compute devices which is a list of device strings and one for each
        # replica. When there are GPUs, replicate operations on these GPUs.
        # Otherwise, place operations on CPU.
        if num_gpus_per_worker > 0:
            self._compute_devices = tuple(
                map("/device:GPU:{}".format, self._visible_gpus)
            )
        else:
            self._compute_devices = (_LOCAL_CPU,)

        self._compute_devices = tuple(
            map(device_util.resolve, self._compute_devices)
        )
        self._canonical_compute_device_set = set(self._compute_devices)

        # If there is only one GPU, put everything on that GPU. Otherwise, place
        # variables on CPU.
        if num_gpus_per_worker == 1:
            assert len(self._compute_devices) == 1
            self._variable_device = _LOCAL_GPU_0
            self._parameter_devices = (_LOCAL_GPU_0,)
        else:
            self._variable_device = _LOCAL_CPU
            self._parameter_devices = (_LOCAL_CPU,)

        self._is_chief = True
        self._cluster_spec = None
        self._task_type = None
        self._task_id = None


class PatchedParameterServerStrategy(ParameterServerStrategy):

    def __init__(self, visible_gpus):
        """Initializes this strategy.
        Args:
        visible_gpus: device IDs of local GPUs to use
        """
        super(ParameterServerStrategy, self).__init__(
            PatchedParameterServerExtended(
                self,
                num_gpus_per_worker=len(visible_gpus),
                visible_gpus=visible_gpus
            )
        )



import tensorflow as tf

from finetune.util.shapes import shape_list


def sample_with_temperature(logits, temperature):
    """Either argmax or random sampling.
    Args:
      logits: a Tensor.
      temperature: a float  0.0=argmax 1.0=random
    Returns:
      a Tensor with one fewer dimension than logits.
    """
    logits_shape = shape_list(logits)
    if temperature == 0.0:
        return tf.argmax(logits, axis=-1)
    else:
        assert temperature > 0.0
        reshaped_logits = tf.reshape(logits, [-1, logits_shape[-1]]) / temperature
        choices = tf.multinomial(reshaped_logits, 1)
        choices = tf.reshape(choices, logits_shape[:-1])
        return choices



import logging
import os
import os.path
import subprocess
import warnings
from collections import namedtuple
from functools import lru_cache

import numpy as np
from nltk.metrics.distance import edit_distance

import finetune
from finetune.errors import FinetuneError
from finetune.base_models import GPTModel, GPT2Model, BERT

LOGGER = logging.getLogger('finetune')


def finetune_model_path(path):
    return os.path.abspath(
        os.path.join(
            os.path.dirname(finetune.__file__),
            'model',
            path
        )
    )


@lru_cache()
def all_gpus(visible_gpus=None):
    """
    Get integer ids of all available GPUs.

    Sample response from nvidia-smi -L:
        GPU 0: GeForce GTX 980 (UUID: GPU-2d683060-957f-d5ad-123c-a5b49b0116d9)
        GPU 1: GeForce GTX 980 (UUID: GPU-7b8496dc-3eaf-8db7-01e7-c4a884f66acf)
        GPU 2: GeForce GTX TITAN X (UUID: GPU-9e01f108-e7de-becd-2589-966dcc1c778f)
    """
    if visible_gpus is not None:
        visible_gpus = [int(gpu) for gpu in visible_gpus]
    try:
        sp = subprocess.Popen(['nvidia-smi', '-L'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        response = sp.communicate()[0]
        gpu_list = response.decode('utf-8').strip().split('\n')
        device_ids = {}
        for i, gpu in enumerate(gpu_list):
            # May be worth logging GPU description
            device_id_str, _, description = gpu.partition(':')
            assert int(device_id_str.split(' ')[-1]) == i
            device_ids[i] = description

        cuda_visible_devices = os.getenv("CUDA_VISIBLE_DEVICES")

        # restricting GPUs based on env vars
        if cuda_visible_devices:
            device_ids = {
                device_id: description
                for device_id, description in device_ids.items()
                if str(device_id) in cuda_visible_devices.split(',')
            }

        # restricting GPUs based on config
        if visible_gpus is not None:
            device_ids = {
                device_id: description
                for device_id, description in device_ids.items()
                if device_id in visible_gpus
            }

        LOGGER.info(" Visible GPUs: {{{}}}".format(
            ", ".join([
                "{}:{}".format(device_id, description.split('(')[0]).strip()
                for device_id, description in device_ids.items()
            ])
        ))

        device_ids = list(device_ids.keys())
    except:
        # Failed to parse out available GPUs properly
        warnings.warn("Failed to find available GPUS.  Falling back to CPU only mode.")
        device_ids = []

    return device_ids

GridSearchable = namedtuple("GridSearchable", "default iterator")


class Settings(dict):
    """
    Model configuration options

    :param batch_size: Number of examples per batch, defaults to `2`.
    :param visible_gpus: List of integer GPU ids to spread out computation across, defaults to all available GPUs.
    :param n_epochs: Number of iterations through training data, defaults to `3`.
    :param random_seed: Random seed to use for repeatability purposes, defaults to `42`.
    :param max_length:  Maximum number of subtokens per sequence. Examples longer than this number will be truncated
        (unless `chunk_long_sequences=True` for SequenceLabeler models). Defaults to `512`.
    :param weight_stddev: Standard deviation of initial weights.  Defaults to `0.02`.
    :param chunk_long_sequences: When True, use a sliding window approach to predict on
        examples that are longer than max length.  Defaults to `False`.
    :param low_memory_mode: When True, only store partial gradients on forward pass
        and recompute remaining gradients incrementally in order to save memory.  Defaults to `False`.
    :param interpolate_pos_embed: Interpolate positional embeddings when `max_length` differs from it's original value of
        `512`. Defaults to `False`.
    :param embed_p_drop: Embedding dropout probability.  Defaults to `0.1`.
    :param attn_p_drop: Attention dropout probability.  Defaults to `0.1`.
    :param resid_p_drop: Residual layer fully connected network dropout probability.  Defaults to `0.1`.
    :param clf_p_drop: Classifier dropout probability.  Defaults to `0.1`.
    :param l2_reg: L2 regularization coefficient. Defaults to `0.01`.
    :param vector_l2: Whether to apply weight decay regularization to vectors (biases, normalization etc..). Defaults to False.
    :param optimizer: Optimizer to use, current options include AdamW or AdamaxW.
    :param b1: Adam b1 parameter.  Defaults to `0.9`.
    :param b2: Adam b2 parameter.  Defaults to `0.999`.
    :param epsilon: Adam epsilon parameter: Defaults to `1e-8`.
    :param lr_schedule: Learning rate schedule -- see `finetune/optimizers.py` for more options.
    :param lr: Learning rate.  Defaults to `6.25e-5`.
    :param lr_warmup: Learning rate warmup (percentage of all batches to warmup for).  Defaults to `0.002`.
    :param max_grad_norm: Clip gradients larger than this norm. Defaults to `1.0`.
    :param accum_steps: Number of updates to accumulate before applying. This is used to simulate a higher batch size.
    :param lm_loss_coef: Language modeling loss coefficient -- a value between `0.0` - `1.0`
        that indicates how to trade off between language modeling loss
        and target model loss.  Usually not beneficial to turn on unless
        dataset size exceeds a few thousand examples.  Defaults to `0.0`.
    :param summarize_grads: Include gradient summary information in tensorboard.  Defaults to `False`.
    :param val_size: Validation set size as a percentage of all training data.  Validation will not be run by default if n_examples < 50.
        If n_examples > 50, defaults to max(5, min(100, 0.05 * n_examples))
    :param val_interval: Evaluate on validation set after `val_interval` batches.
        Defaults to 4 * val_size / batch_size to ensure that too much time is not spent on validation.
    :param lm_temp: Language model temperature -- a value of `0.0` corresponds to greedy maximum likelihood predictions
        while a value of `1.0` corresponds to random predictions. Defaults to `0.2`.
    :param seq_num_heads: Number of attention heads of final attention layer. Defaults to `16`.
    :param subtoken_predictions: Return predictions at subtoken granularity or token granularity?  Defaults to `False`.
    :param multi_label_sequences: Use a multi-labeling approach to sequence labeling to allow overlapping labels.
    :param multi_label_threshold: Threshold of sigmoid unit in multi label classifier.
        Can be increased or lowered to trade off precision / recall. Defaults to `0.5`.
    :param autosave_path: Save current best model (as measured by validation loss) to this location. Defaults to `None`.
    :param tensorboard_folder: Directory for tensorboard logs. Tensorboard logs will not be written
        unless tensorboard_folder is explicitly provided. Defaults to `None`.
    :param log_device_placement: Log which device each operation is placed on for debugging purposes.  Defaults to `False`.
    :param allow_soft_placement: Allow tf to allocate an operation to a different device if a device is unavailable.  Defaults to `True`.
    :param save_adam_vars: Save adam parameters when calling `model.save()`.  Defaults to `True`.
    :param num_layers_trained: How many layers to finetune.  Specifying a value less than 12 will train layers starting from model output. Defaults to `12`.
    :param train_embeddings: Should embedding layer be finetuned? Defaults to `True`.
    :param class_weights: One of 'log', 'linear', or 'sqrt'. Auto-scales gradient updates based on class frequency.  Can also be a dictionary that maps from true class name to loss coefficient. Defaults to `None`.
    :param oversample: Should rare classes be oversampled?  Defaults to `False`.
    :param params_device: Which device should gradient updates be aggregated on?
        If you are using a single GPU and have more than 4Gb of GPU memory you should set this to GPU PCI number (0, 1, 2, etc.). Defaults to `"cpu"`.
    :param eval_acc: if True, calculates accuracy and writes it to the tensorboard summary files for valudation runs.
    :param save_dtype: specifies what precision to save model weights with.  Defaults to `np.float32`.
    :param regression_loss: the loss to use for regression models. One of `L1` or `L2`, defaults to `L2`.
    :param prefit_init: if True, fit target model weigths before finetuning the entire model. Defaults to `False`.
    :param debugging_logs: if True, output tensorflow logs and turn off TQDM logging. Defaults to `False`.
    :param val_set: Where it is neccessary to use an explicit validation set, provide it here as a tuple (text, labels)
    :param per_process_gpu_memory_fraction: fraction of the overall amount of memory that each visible GPU should be allocated, defaults to `1.0`.
    """

    def get_grid_searchable(self):
        return self.grid_searchable

    def __init__(self, **kwargs):
        super().__init__()
        self.grid_searchable = {}
        for key, value in kwargs.items():
            self[key] = value

    def __getattr__(self, attr):
        if attr.startswith('__'):
            raise AttributeError

        if attr == "base_model_path":
            full_path = finetune_model_path(self["base_model_path"])
            if os.path.exists(full_path):
                return full_path

        return self[attr]

    def __setitem__(self, key, value):
        if isinstance(value, GridSearchable):
            self.grid_searchable[key] = value.iterator
            value = value.default
        return super().__setitem__(key, value)

    def __setattr__(self, k, v):
        return self.__setitem__(k, v)

    __delattr__ = dict.__delitem__


def did_you_mean(keyword, keyword_pool):
    candidates = list(keyword_pool)
    closest_match_idx = np.argmin([
        edit_distance(keyword, candidate) for candidate in candidates
    ])
    return candidates[closest_match_idx]


def assert_valid_config(**kwargs):
    expected_keys = set(get_default_config().keys())
    for kwarg in kwargs:
        if kwarg not in expected_keys:
            raise FinetuneError(
                "Unexpected setting configuration: `{}` is an invalid keyword. "
                "Did you mean `{}`?".format(kwarg, did_you_mean(kwarg, expected_keys))
            )


def get_default_config():
    """
    Gets a config object containing all the default parameters for each variant of the model.

    :return: Config object.
    """
    settings = Settings(
        # General Settings
        low_memory_mode=False,
        interpolate_pos_embed=False,
        save_adam_vars=True,
        shuffle_buffer_size=100,
        dataset_size=None,
        batch_size=2,
        visible_gpus=None, # defaults to all available
        n_epochs=GridSearchable(3, [1, 2, 3, 4]),
        seed=42,
        max_length=512,
        weight_stddev=0.02,
        save_dtype=None,
        val_set=None,
        per_process_gpu_memory_fraction=0.95,

        # Regularization
        embed_p_drop=0.1,
        attn_p_drop=0.1,
        resid_p_drop=0.1,
        clf_p_drop=0.1,
        l2_reg=GridSearchable(0.01, [0.0, 0.1, 0.01, 0.001]),
        vector_l2=False,

        # Early Stopping and Validation
        autosave_path=None,
        keep_best_model=False,
        early_stopping_steps=None,
        min_secs_between_eval=60,
        eval_acc=False,
        val_size=None,
        val_interval=None,

        # Debugging
        log_device_placement=False,
        soft_device_placement=True,
        tensorboard_folder=None,
        summarize_grads=False,
        debugging_logs=False,

        # Partial Fitting
        num_layers_trained=12,
        train_embeddings=True,

        # Class Imbalance
        class_weights=None,
        oversample=False,
        params_device="cpu",

        # Optimization Params
        optimizer="AdamW",
        b1=0.9,
        b2=0.999,
        epsilon=1e-8,
        lr_schedule='warmup_linear',
        lr=GridSearchable(6.25e-5, [6.25e-4, 6.25e-5, 6.25e-6]),
        lr_warmup=0.002,
        max_grad_norm=1.0,
        prefit_init=False,
        accum_steps=1,

        # MTL
        tasks=None,
        dont_optimize_zero_gradients=False,

        # Language Model Settings
        lm_loss_coef=0.0,
        lm_temp=0.2,

        # Sequence Labeling
        seq_num_heads=16,
        pad_token="<PAD>",
        pad_idx=None,
        subtoken_predictions=False,
        multi_label_sequences=False,
        multi_label_threshold=0.5,
        chunk_long_sequences=False,

        # Regression Params
        regression_loss="L2",

        # Association Params
        viable_edges=None,
        association_types=None,
        assocation_loss_weight=100.0,

        # Location of model weights
        base_model=GPTModel,
        base_model_path=None,

        # Possible `SourceModel` specific settings
        n_heads=None,
        n_layer=None,
        act_fn=None,
        n_embed=None,

        # for TextCNN SourceModel only
        kernel_sizes=None,
        num_filters_per_size=None,
        n_embed_featurizer=None, # needed because the dimensions CNN output are different from the embedding dimensions

        # BERT only
        bert_intermediate_size=None
    )
    return settings


def get_config(**kwargs):
    """
    Gets a config object containing all the default parameters for each variant of the model.

    :param **kwargs: Keyword arguments to override default values.
    :return: Config object.    """
    assert_valid_config(**kwargs)
    config = get_default_config()
    config.base_model = kwargs.get('base_model', config.base_model)
    config.update(config.base_model.settings)
    config.update(kwargs)
    return config


def cpu_config(**kwargs):
    config = get_config(**kwargs)
    config.visible_gpus = []
    config.update(kwargs)
    return config



import numpy as np
from sklearn.preprocessing import LabelEncoder, MultiLabelBinarizer, OrdinalEncoder
from abc import ABCMeta


class BaseEncoder(metaclass=ABCMeta):
    @property
    def target_labels(self):
        return getattr(self, 'classes_', None)

    @property
    def target_dim(self):
        return len(self.target_labels) if self.target_labels is not None else None


class RegressionEncoder(BaseEncoder):
    def __init__(self):
        self.num_outputs = None

    def fit(self, x):
        self.fit_transform(x)
        return self

    def transform(self, x):
        output = np.array(x)
        rank = len(output.shape)
        if rank == 1:
            return np.expand_dims(output, 1)  # for single output value regression.
        if rank == 2:
            return output
        raise ValueError("Unresolvable shape: {}. Must be able to fit a format [batch, n_outputs]".format(output.shape))

    def fit_transform(self, x):
        output = self.transform(x)
        self.num_outputs = output.shape[1]
        return output

    def inverse_transform(self, y):
        y = np.array(y)
        if y.shape[1] == 1:
            return np.squeeze(y, 1)
        else:
            return y

    @property
    def target_dim(self):
        return self.num_outputs

    @property
    def target_labels(self):
        raise ValueError


class OneHotLabelEncoder(LabelEncoder, BaseEncoder):

    def _make_one_hot(self, labels):
        output = np.zeros([len(labels), len(self.classes_)], dtype=np.float)
        output[np.arange(len(labels)), labels] = 1
        return output

    def fit_transform(self, y):
        labels = super().fit_transform(y)
        return self._make_one_hot(labels)

    def transform(self, y):
        labels = super().transform(y)
        return self._make_one_hot(labels)

class OrdinalRegressionEncoder(OrdinalEncoder, BaseEncoder):

    def __init__(self):
        self.num_outputs = None
        super().__init__()

    def _force_2d(self, x):
        return np.array(x, dtype=np.int32).reshape(-1, 1)

    def fit(self, x):
        super().fit(self._force_2d(x))
        self.num_outputs = len(self.categories_[0]) - 1
        return self

    def transform(self, x):
        labels = super().transform(self._force_2d(x)).astype(np.int32)
        labels = self.rank_to_one_hot(labels)
        return labels

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)

    def rank_to_one_hot(self, x):
        # changes a one-variable rank into an array of 1s and 0s defining the target output of each threshold
        one_hot = np.zeros((len(x), self.num_outputs), dtype=np.float32)
        for i, (rank,) in enumerate(x):
            one_hot[i, :rank] = 1
        return one_hot

    def inverse_transform(self, y):
        y = np.array(y)
        y = y > 0.5
        rank = np.sum(y, axis=1)
        rank = np.expand_dims(rank, 1)
        y = super().inverse_transform(rank)
        y = np.squeeze(y)
        return y

    @property
    def target_dim(self):
        return self.num_outputs

    @property
    def target_labels(self):
        raise ValueError

class SequenceLabelingEncoder(LabelEncoder, BaseEncoder):
    pass


class SequenceMultiLabelingEncoder(MultiLabelBinarizer, BaseEncoder):
    pass


class MultilabelClassificationEncoder(MultiLabelBinarizer, BaseEncoder):
    pass


class IDEncoder(BaseEncoder):

    def __init__(self):
        self.classes_ = [0]

    def transform(self, x):
        return x

    def fit(self, x):
        return x

    def fit_transform(self, x):
        return x

    def inverse_transform(self, x):
        return x



import os
import logging
from pathlib import Path
import json

import numpy as np

from sklearn.model_selection import train_test_split

from finetune import MultiTask, Classifier, Comparison
from finetune.datasets.stanford_sentiment_treebank import StanfordSentimentTreebank
from finetune.datasets.quora_similarity import QuoraDuplicate
from finetune.datasets.reuters import Reuters
logging.basicConfig(level=logging.DEBUG)

SST_FILENAME = "SST-binary.csv"
DATA_PATH = os.path.join('Data', 'Classify', SST_FILENAME)
CHECKSUM = "02136b7176f44ff8bec6db2665fc769a"

if __name__ == "__main__":
    # Train and evaluate on SST
    dataset = StanfordSentimentTreebank(nrows=1000).dataframe
    q_dataset = QuoraDuplicate(nrows=1000).dataframe
    r_dataset = Reuters().dataframe
    r_dataset['annotations'] = [json.loads(annotation) for annotation in r_dataset['annotations']]
    r_trainX, r_testX, r_trainY, r_testY = train_test_split(
        r_dataset.texts.values,
        r_dataset.annotations.values,
        test_size=0.3,
        random_state=42
    )
    
    model = MultiTask(tasks={
        "sst": Classifier,
        "qqp": Comparison
    }, debugging_logs=True, lr=6.25e-6, n_epochs=2)

    q_trainX1, q_testX1, q_trainX2, q_testX2, q_trainY, q_testY = train_test_split(
        q_dataset.Text1.values, q_dataset.Text2.values, q_dataset.Target.values,
        test_size=0.3, random_state=42
    )
    trainX, testX, trainY, testY = train_test_split(
        dataset.Text.values, dataset.Target.values, test_size=0.3,  random_state=42
    )
    model.fit(
        {
            "sst": trainX,
            "qqp": list(zip(q_trainX1, q_trainX2)),
        },
        {
            "sst": trainY,
            "qqp": q_trainY,
        }
    )
    model.create_base_model("mtl_sst_qqp_test.jl", exists_ok=True)
    model = Classifier(base_model_path="mtl_sst_qqp_test.jl")
    model.fit(trainX, trainY)
    accuracy = np.mean(model.predict(testX) == testY)
    print('Test Accuracy: {:0.2f}'.format(accuracy))



import tensorflow as tf

from finetune.base import BaseModel
from finetune.encoding.target_encoders import OrdinalRegressionEncoder
from finetune.nn.target_blocks import ordinal_regressor
from finetune.input_pipeline import BasePipeline
from finetune.target_models.comparison_regressor import ComparisonRegressionPipeline


class OrdinalRegressionPipeline(BasePipeline):
    def _target_encoder(self):
        return OrdinalRegressionEncoder()


class ComparisonOrdinalRegressionPipeline(ComparisonRegressionPipeline):
    def _target_encoder(self):
        return OrdinalRegressionEncoder()


class OrdinalRegressor(BaseModel):
    """
    Classifies a document into two or more ordered categories.

    For a full list of configuration options, see `finetune.config`.

    :param config: A config object generated by `finetune.config.get_config` or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    def __init__(self, shared_threshold_weights=True, **kwargs):
        super().__init__(**kwargs)
        self.config.shared_threshold_weights = shared_threshold_weights

    def _get_input_pipeline(self):
        return OrdinalRegressionPipeline(self.config)

    def featurize(self, X):
        """
        Embeds inputs in learned feature space. Can be called before or after calling :meth:`finetune`.

        :param X: list or array of text to embed.
        :returns: np.array of features of shape (n_examples, embedding_size).
        """
        return self._featurize(X)

    def predict(self, X):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.

        :param X: list or array of text to embed.
        :returns: list of class labels.
        """
        return super().predict(X).tolist()

    def predict_proba(self, X):
        """
        Produces a probability distribution over classes for each example in X.

        :param X: list or array of text to embed.
        :returns: list of dictionaries.  Each dictionary maps from a class label to its assigned class probability.
        """
        raise AttributeError("`Regressor` model does not support `predict_proba`.")

    def finetune(self, X, Y=None, batch_size=None):
        """
        :param X: list or array of text.
        :param Y: floating point targets
        :param batch_size: integer number of examples per batch. When N_GPUS > 1, this number
                           corresponds to the number of training examples provided to each GPU.
        """
        return super().finetune(X, Y=Y, batch_size=batch_size)

    @staticmethod
    def _target_model(config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs):
        return ordinal_regressor(
            hidden=featurizer_state['features'],
            targets=targets,
            n_targets=n_outputs,
            config=config,
            train=train,
            reuse=reuse,
            shared_threshold_weights=config.shared_threshold_weights,
            **kwargs
        )

    def _predict_op(self, logits, **kwargs):
        return logits

    def _predict_proba_op(self, logits, **kwargs):
        return logits


class ComparisonOrdinalRegressor(OrdinalRegressor):
    """
    Compares two documents and classifies into two or more ordered categories.

    For a full list of configuration options, see `finetune.config`.

    :param config: A config object generated by `finetune.config.get_config` or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    def _get_input_pipeline(self):
        return ComparisonOrdinalRegressionPipeline(self.config)

    @staticmethod
    def _target_model(config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs):
        featurizer_state["sequence_features"] = tf.abs(tf.reduce_sum(featurizer_state["sequence_features"], 1))
        featurizer_state["features"] = tf.abs(tf.reduce_sum(featurizer_state["features"], 1))
        return ordinal_regressor(
            hidden=featurizer_state['features'],
            targets=targets,
            n_targets=n_outputs,
            config=config,
            shared_threshold_weights=config.shared_threshold_weights,
            train=train,
            reuse=reuse,
            **kwargs
        )




import json
import os
import codecs
import random
import requests
from pathlib import Path
import finetune
from finetune import MultipleChoice
from sklearn.model_selection import train_test_split
import numpy as np

QnA = "QandA"
DATA_PATH = os.path.join('Data', 'QA', QnA)
CHECKSUM = ""


def download_data():
    base_url = "https://raw.githubusercontent.com/uberspot/OpenTriviaQA/master/categories/{}"

    file_list = [
        "animals",
        "general",
        "geography",
        "history",
        "literature",
        "people"
    ]

    for filename in file_list:
        folder = DATA_PATH
        if not os.path.exists(folder):
            os.mkdir(folder)

        local_filepath = os.path.join(folder, filename)

        print(base_url.format(filename))
        data = requests.get(base_url.format(filename)).content
        fd = open(local_filepath, 'wb')
        fd.write(data)
        fd.close()

def parse_file(filename):
    with codecs.open(filename, encoding='ISO-8859-2') as file:
        file_content = file.readlines()
    questions = []
    last_question = None
    inside_question = False
    correct_answer = None
    for line in file_content:
        line = line.rstrip()
        if line == "":
            continue
        elif line[0] == "#":
            if last_question is not None:
                last_question["correct_answer"] = last_question["answers"].index(correct_answer)
                questions.append(last_question)

            last_question = {
                "content": line[3:],
                "answers": [],
            }
            inside_question = True
        elif line[0] == "^":
            correct_answer = line[2:]
            inside_question = False
        elif line[0] == "A" or line[0] == "B" or line[0] == "C" or line[0] == "D":
            last_question["answers"].append(line[2:])
        elif inside_question:
            last_question["content"] += "\n" + line
        elif line.strip() != "":
            raise Exception("Can't parse line in file {0}: {1}".format(filename, line))

    if last_question is not None:
        last_question["correct_answer"] = last_question["answers"].index(correct_answer)
        questions.append(last_question)
    return questions

def get_dataset(nrows=None):
    download_data()
    data = []
    files = os.listdir(DATA_PATH)
    for file in files:
        if not file.startswith('.'):
            data.extend(parse_file(os.path.join(DATA_PATH, file)))
    random.shuffle(data)
    return data[:nrows]


if __name__ == "__main__":
    print("DISCLAIMER: THIS DATASET IS NOT WELL SUITED TO THIS MODEL AND DOES NOT ACHIEVE RESULTS MUCH BETTER THAN "
          "RANDOM CHANCE. IT IS HERE FOR API DEMO PURPOSES DUE TO LICENCES OF AVAILABLE DATASETS.")

    data = get_dataset()

    model = MultipleChoice(n_epochs=1)
    train_data, test_data = train_test_split(data, test_size=0.3, random_state=42)

    train_qs = []
    train_ans_correct = []
    train_ans = []
    for item in train_data:
        answers = item["answers"]
        correct_ans_idx = item["correct_answer"]
        train_ans_correct.append(answers[correct_ans_idx])
        train_qs.append(item["content"])
        train_ans.append(answers)

    test_qs = []
    test_ans_all = []
    test_ans_correct = []
    for item in test_data:
        answers = item["answers"]
        correct_ans_idx = item["correct_answer"]
        test_ans_correct.append(answers[correct_ans_idx])
        test_qs.append(item["content"])
        test_ans_all.append(answers)

    model.fit(train_qs, train_ans, train_ans_correct)
    print(list(zip(model.predict(test_qs, test_ans_all), test_qs)))

    accuracy = np.mean([p == t for p, t in zip(model.predict(test_qs, test_ans_all), test_ans_correct)])

    print('Test Accuracy: {:0.2f}'.format(accuracy))






"""
IndicoAPI setup
"""
import os
from sys import version_info
from setuptools import setup, find_packages


REQUIREMENTS = [
    "pandas>=0.23.1",
    "tqdm>=4.0.0",
    "numpy>=1.13.0",
    "scipy>=1.1.0",
    "scikit-learn>=0.18.0",
    "ftfy>=4.4.0",
    "spacy>=2.0.0",
    "msgpack-numpy==0.4.1",
    "h5py>=2.8.0",
    "joblib>=0.12.0",
    "bs4>=0.0.1",
    "imblearn>=0.0",
    "nltk>=3.2.4",
    "regex>=2019.03.12",
    "lxml>=4.3.3"
]

setup(
    name="finetune",
    packages=find_packages(),
    version="0.6.4",
    install_requires=REQUIREMENTS,
    extras_require={
        "tf": ["tensorflow>=1.13.0"],
        "tf_gpu": ["tensorflow-gpu>=1.13.0"],
    },
    include_package_data=False
)



import math
import tensorflow as tf
from itertools import zip_longest


def warmup_cosine(x, warmup=0.002, *args):
    s = tf.cast(x <= warmup, tf.float32)
    return s*(x/warmup) + (1-s)*(0.5 * (1 + tf.cos(math.pi * x)))


def warmup_constant(x, warmup=0.002, *args):
    s = tf.cast(x <= warmup, tf.float32)
    return s*(x/warmup) + (1-s)*1


def warmup_linear(x, warmup=0.002, *args):
    s = tf.cast(x <= warmup, tf.float32)
    return (s*(x/warmup) + (1-s))*(1-x)

schedules = {
    'warmup_cosine': warmup_cosine,
    'warmup_constant': warmup_constant,
    'warmup_linear': warmup_linear,
    'none': lambda x, *args, **kwargs: x
}


import os
import urllib
import urllib.request
from urllib.parse import urljoin
from pathlib import Path

import finetune


def download_data_if_required():
    """ Pulls the pre-trained model weights from Github if required. """
    github_base_url = "https://raw.githubusercontent.com/IndicoDataSolutions/finetune/master/finetune/model/"
    s3_base_url = "https://s3.amazonaws.com/bendropbox/"
    gpt2_base_url = "https://s3.amazonaws.com/bendropbox/gpt2/"
    bert_base_url = "https://s3.amazonaws.com/bendropbox/bert/"
    finetune_base_folder = os.path.dirname(finetune.__file__)
    file_list = [
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt', 'encoder.json'),
            'url': urljoin(github_base_url, "gpt/encoder.json")
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt', 'vocab.bpe'),
            'url': urljoin(github_base_url, "gpt/vocab.bpe")
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt', 'model-lg.jl'),
            'url': urljoin(s3_base_url, "Base_model.jl")
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt', 'model-sm.jl'),
            'url': urljoin(s3_base_url, "SmallBaseModel.jl")
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt2', 'encoder.json'),
            'url': urljoin(gpt2_base_url, 'encoder.json')
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt2', 'vocab.bpe'),
            'url': urljoin(gpt2_base_url, 'vocab.bpe')
        },
        {
            'file': os.path.join(finetune_base_folder, 'model', 'gpt2', 'model-sm.jl'),
            'url': urljoin(gpt2_base_url, 'model-sm.jl')
        }
    ]

    bert_file_list = [
        {
            'file': os.path.join(finetune_base_folder, 'model', 'bert', file),
            'url': urljoin(bert_base_url, file)
        } for file in [
            "bert_large_cased.jl", "bert_small_cased.jl", "bert_small_multi_cased.jl",
            "vocab.txt", "vocab_large.txt", "vocab_multi.txt"
        ]
    ]

    for file_obj in file_list + bert_file_list:
        path = Path(file_obj['file'])
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            print("Downloading: {}".format(path.name))
            data = urllib.request.urlopen(file_obj['url']).read()
            with path.open('wb') as f:
                f.write(data)



import tensorflow as tf


def get_grad_accumulation_optimizer(optimizer_class, accum_steps):
    """
    Adds gradient accumulation to an Optimizer.

    Unlike most optimizers, this will create a global step if one does not already exist.
    This optimizer also increments this global step automatically, whether a step is passed or not.

    :param optimizer_class: A subclass of tf.train.Optimizer to add gradient accumulation too.
    :param accum_steps: An int value determining how many gradients to accumulate before performing an optimizer step.

    :return: A new Optimizer, with gradient accumulation.
    """

    class GradAccumulationOptimizer(optimizer_class):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def apply_gradients(self, grads_and_vars, global_step=None, *args, **kwargs):
            add_gradients_ops = []
            accumulation_vars = []
            grads_and_accumulated_vars = []
            for g, v in grads_and_vars:
                accum_grad = tf.get_variable(
                    name=g.name[:-2] + "_acc",
                    shape=g.shape,
                    dtype=g.dtype,
                    initializer=tf.constant_initializer(0),
                    use_resource=True,
                    trainable=False
                )
                add_gradients_ops.append(accum_grad.assign_add(g))
                accumulation_vars.append(accum_grad)

                grads_and_accumulated_vars.append((accum_grad, v))

            global_step = global_step if global_step is not None else tf.train.get_or_create_global_step()

            with tf.control_dependencies(add_gradients_ops):
                def apply_grads():
                    with tf.control_dependencies([
                        super(GradAccumulationOptimizer, self).apply_gradients(
                            grads_and_accumulated_vars, global_step=global_step, *args, **kwargs
                        )
                    ]):
                        apply_grads_op = tf.group(*[g.assign(g.initial_value) for g in accumulation_vars])
                    return apply_grads_op

                return tf.cond(
                    tf.equal(global_step % accum_steps, accum_steps - 1),
                    true_fn=apply_grads,
                    false_fn=lambda: tf.group(global_step.assign_add(1))
                )

    return GradAccumulationOptimizer



from collections import defaultdict

from sklearn.metrics import accuracy_score, recall_score, precision_score
import numpy as np

from finetune.encoding.input_encoder import NLP


def _convert_to_token_list(annotations, doc_idx=None):
    tokens = []

    for annotation in annotations:
        start_idx = annotation.get('start')
        tokens.extend([
            {
                'start': start_idx + token.idx,
                'end': start_idx + token.idx + len(token.text),
                'text': token.text,
                'label': annotation.get('label'),
                'doc_idx': doc_idx
            }
            for token in NLP(annotation.get('text'))
        ])

    return tokens


def sequence_labeling_token_counts(true, predicted):
    """
    Return FP, FN, and TP counts
    """

    unique_classes = set([seq['label'] for seqs in true for seq in seqs])

    d = {
        cls_: {
            'false_positives': [],
            'false_negatives': [],
            'correct': []
        }
        for cls_ in unique_classes
    }
    
    for i, (true_list, pred_list) in enumerate(zip(true, predicted)):
        true_tokens = _convert_to_token_list(true_list, doc_idx=i)
        pred_tokens = _convert_to_token_list(pred_list, doc_idx=i)

        # correct + false negatives
        for true_token in true_tokens:
            for pred_token in pred_tokens:
                if (pred_token['start'] == true_token['start'] and 
                    pred_token['end'] == true_token['end']):

                    if pred_token['label'] == true_token['label']:
                        d[true_token['label']]['correct'].append(true_token)
                    else:
                        d[true_token['label']]['false_negatives'].append(true_token)
                        d[pred_token['label']]['false_positives'].append(pred_token)
                    
                    break
            else:
                d[true_token['label']]['false_negatives'].append(true_token)

        # false positives
        for pred_token in pred_tokens:
            for true_token in true_tokens:
                if (pred_token['start'] == true_token['start'] and 
                    pred_token['end'] == true_token['end']):
                    break
            else:
                d[pred_token['label']]['false_positives'].append(pred_token)
    
    return d


def seq_recall(true, predicted, count_fn):
    class_counts = count_fn(true, predicted)
    results = {}
    for cls_, counts in class_counts.items():
        FN = len(counts['false_negatives'])
        TP = len(counts['correct'])
        try:
            results[cls_] = TP / float(FN + TP)
        except ZeroDivisionError: 
            results[cls_] = 0.
    return results


def seq_precision(true, predicted, count_fn):
    class_counts = count_fn(true, predicted)
    results = {}
    for cls_, counts in class_counts.items():
        FP = len(counts['false_positives'])
        TP = len(counts['correct'])
        try:
            results[cls_] = TP / float(FP + TP)
        except ZeroDivisionError:
            results[cls_] = 0.
    return results

def micro_f1(true, predicted, count_fn):
    class_counts = count_fn(true, predicted)
    TP, FP, FN = 0, 0, 0
    for cls_, counts in class_counts.items():
        FN += len(counts['false_negatives'])
        TP += len(counts['correct'])
        FP += len(counts['false_positives'])
    recall = TP/float(FN + TP)
    precision = TP / float(FP + TP)
    f1 = 2 * (recall * precision) / (recall + precision)
    return f1

def sequence_labeling_token_precision(true, predicted):
    """
    Token level precision
    """
    return seq_precision(true, predicted, count_fn=sequence_labeling_token_counts)


def sequence_labeling_token_recall(true, predicted):
    """
    Token level recall
    """
    return seq_recall(true, predicted, count_fn=sequence_labeling_token_counts)

def sequence_labeling_micro_token_f1(true, predicted):
    """
    Token level F1
    """
    return micro_f1(true, predicted, count_fn=sequence_labeling_token_counts)

def sequences_overlap(true_seq, pred_seq):
    """
    Boolean return value indicates whether or not seqs overlap
    """
    start_contained = (pred_seq['start'] < true_seq['end'] and pred_seq['start'] >= true_seq['start'])
    end_contained = (pred_seq['end'] > true_seq['start'] and pred_seq['end'] <= true_seq['end'])
    return start_contained or end_contained


def sequence_labeling_overlaps(true, predicted):
    """
    Return FP, FN, and TP counts
    """
    unique_classes = set([annotation['label'] for annotations in true for annotation in annotations])

    d = {
        cls_: {
            'false_positives': [],
            'false_negatives': [],
            'correct': []
        }
        for cls_ in unique_classes
    }

    for i, (true_annotations, predicted_annotations) in enumerate(zip(true, predicted)):
        # add doc idx to make verification easier
        for annotations in [true_annotations, predicted_annotations]:
            for annotation in annotations:
                annotation['doc_idx'] = i
        
        for true_annotation in true_annotations:
            for pred_annotation in predicted_annotations:
                if sequences_overlap(true_annotation, pred_annotation):
                    if pred_annotation['label'] == true_annotation['label']:
                        d[true_annotation['label']]['correct'].append(true_annotation)
                    else:
                        d[true_annotation['label']]['false_negatives'].append(true_annotation)
                        d[pred_annotation['label']]['false_positives'].append(pred_annotation)
                    break
            else:
                d[true_annotation['label']]['false_negatives'].append(true_annotation)

        for pred_annotation in predicted_annotations:
            for true_annotation in true_annotations:
                if (sequences_overlap(true_annotation, pred_annotation) and
                        true_annotation['label'] == pred_annotation['label']):
                    break
            else:
                d[pred_annotation['label']]['false_positives'].append(pred_annotation)

    return d


def sequence_labeling_overlap_precision(true, predicted):
    """
    Sequence overlap precision
    """
    return seq_precision(true, predicted, count_fn=sequence_labeling_overlaps)


def sequence_labeling_overlap_recall(true, predicted):
    """
    Sequence overlap recall
    """
    return seq_recall(true, predicted, count_fn=sequence_labeling_overlaps)


def annotation_report(y_true, y_pred, labels=None, target_names=None, sample_weight=None, digits=2, width=20):
    # Adaptation of https://github.com/scikit-learn/scikit-learn/blob/f0ab589f/sklearn/metrics/classification.py#L1363
    token_precision = sequence_labeling_token_precision(y_true, y_pred)
    token_recall = sequence_labeling_token_recall(y_true, y_pred)
    overlap_precision = sequence_labeling_overlap_precision(y_true, y_pred)
    overlap_recall = sequence_labeling_overlap_recall(y_true, y_pred)

    count_dict = defaultdict(int)
    for annotation_seq in y_true:
        for annotation in annotation_seq:
            count_dict[annotation['label']] += 1

    seqs = [token_precision, token_recall, overlap_precision, overlap_recall, dict(count_dict)]
    labels = set(token_precision.keys()) | set(token_recall.keys())
    target_names = [u'%s' % l for l in labels]
    counts = [count_dict.get(target_name) for target_name in target_names]

    last_line_heading = 'Weighted Summary'
    headers = ["token_precision", "token_recall", "overlap_precision", "overlap_recall", "support"]
    head_fmt = u'{:>{width}s} ' + u' {:>{width}}' * len(headers)
    report = head_fmt.format(u'', *headers, width=width)
    report += u'\n\n'
    row_fmt = u'{:>{width}s} ' + u' {:>{width}.{digits}f}' * 4 + u' {:>{width}}' '\n'
    seqs = [
        [seq.get(target_name) for target_name in target_names]
        for seq in seqs
    ]
    rows = zip(target_names, *seqs)
    for row in rows:
        report += row_fmt.format(*row, width=width, digits=digits)

    report += u'\n'
    averages = [np.average(seq, weights=counts) for seq in seqs[:-1]] + [np.sum(seqs[-1])]
    report += row_fmt.format(last_line_heading, *averages, width=width, digits=digits)
    return report



from finetune.base import PredictMode
from finetune.target_models.classifier import Classifier


class LanguageModel(Classifier):
    """ 
    A Language Model for Finetune

    :param config: A :py:class:`finetune.config.Settings` object or None (for default config).
    :param \**kwargs: key-value pairs of config items to override.
    """

    def predict(self, X):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.

        :param X: list or array of text to embed.
        :returns: Perplexities of each of the input sentences.
        """
        return self._inference(X, predict_keys=[PredictMode.LM_PERPLEXITY])

    def predict_proba(self, X):
        raise ValueError("Predict Proba is not defined for the language model")

    def finetune(self, X, Y=None, batch_size=None):
        """
        :param X: list or array of text.
        :param Y: Not used.
        :param batch_size: integer number of examples per batch. When N_GPUS > 1, this number
                           corresponds to the number of training examples provided to each GPU.
        """
        if Y:
            raise ValueError("No targets should be provided for the language model")
        return super().finetune(X, Y=None, batch_size=batch_size)



import os
import logging
from pathlib import Path
import hashlib

import requests
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split

from finetune import MultiFieldClassifier
from finetune.datasets import Dataset

logging.basicConfig(level=logging.DEBUG)

FILENAME = "multinli.dev.csv"
DATA_PATH = os.path.join('Data', 'Entailment', FILENAME)
CHECKSUM = "4837f671a2ee1042f3d308de5352b58e"


class MultiNLI(Dataset):

    def __init__(self, filename=None, **kwargs):
        super().__init__(filename=(filename or DATA_PATH), **kwargs)


    @property
    def md5(self):
        return CHECKSUM

    def download(self):
        """
        Download Stanford Sentiment Treebank to data directory
        """
        path = Path(self.filename)
        path.parent.mkdir(parents=True, exist_ok=True)

        remote_url = "https://s3.amazonaws.com/enso-data/multinli.dev.csv"

        response = requests.get(remote_url)
        open(DATA_PATH, 'wb').write(response.content)



if __name__ == "__main__":
    # Train and evaluate on MultiNLI
    dataset = MultiNLI(nrows=1000).dataframe
    model = MultiFieldClassifier()
    trainX1, testX1, trainX2, testX2, trainY, testY = train_test_split(
        dataset.x1, dataset.x2, dataset.target, test_size=0.3, random_state=42
    )
    model.fit(list(zip(trainX1, trainX2)), Y=trainY)
    accuracy = np.mean(model.predict(list(zip(testX1, testX2))) == testY)
    print('Test Accuracy: {:0.2f}'.format(accuracy))



import os
from finetune.base_models import SourceModel
from finetune.base_models.gpt.encoder import GPTEncoder
from finetune.base_models.textcnn.featurizer import textcnn_featurizer


class TextCNNModel(SourceModel):
    is_bidirectional = False
    encoder = GPTEncoder
    featurizer = textcnn_featurizer
    kernel_sizes = [2, 3, 4]
    settings = {
        'n_embed_featurizer': 768,
        'n_layer': 1,
        'num_layers_trained': 1,
        'kernel_sizes': kernel_sizes,
        'num_filters_per_size': 2,
        'n_embed': len(kernel_sizes) * 2,
        'act_fn': "gelu",
        'train_embeddings': False,
        'lr': .01,
        'seq_num_heads': len(kernel_sizes) * 2,
        'base_model_path': os.path.join("gpt", "model-lg.jl"),
    }



import tensorflow as tf
from finetune.base_models.bert.modeling import BertConfig, BertModel


def bert_featurizer(X, encoder, config, train=False, reuse=None):
    """
    The transformer element of the finetuning model. Maps from tokens ids to a dense, embedding of the sequence.

    :param X: A tensor of token indexes with shape [batch_size, sequence_length, token_idx]
    :param encoder: A TextEncoder object.
    :param config: A config object, containing all parameters for the featurizer.
    :param train: If this flag is true, dropout and losses are added to the graph.
    :param reuse: Should reuse be set within this scope.
    :return: A dict containing;
        embed_weights: the word embedding matrix.
        features: The output of the featurizer_final state.
        sequence_features: The output of the featurizer at each timestep.
    """

    bert_config = BertConfig(
        vocab_size=encoder.vocab_size,
        hidden_size=config.n_embed,
        num_hidden_layers=config.n_layer,
        num_attention_heads=config.n_heads,
        intermediate_size=config.bert_intermediate_size,
        hidden_act=config.act_fn,
        hidden_dropout_prob=config.resid_p_drop,
        attention_probs_dropout_prob=config.attn_p_drop,
        max_position_embeddings=config.max_length,
        type_vocab_size=2,
        initializer_range=config.weight_stddev
    )

    initial_shape = tf.shape(X)
    X = tf.reshape(X, shape=tf.concat(([-1], initial_shape[-2:]), 0))
    # To fit the interface of finetune we are going to compute the mask and type id at runtime.
    input_ids = X[:, :, 0]  # slice off pos-embed ids.
    delimiters = tf.cast(tf.equal(input_ids, encoder.delimiter), tf.int32)
    token_type_ids = tf.cumsum(delimiters, exclusive=True, axis=1)

    seq_length = tf.shape(delimiters)[1]

    lengths = tf.argmax(
        tf.cast(delimiters, tf.float32) *
        tf.expand_dims(tf.range(tf.cast(seq_length, tf.float32), dtype=tf.float32), 0),
        axis=1
    )

    mask = tf.sequence_mask(lengths, maxlen=seq_length, dtype=tf.float32)
    if config.num_layers_trained not in [config.n_layer, 0]:
        raise ValueError("Bert base model does not support num_layers_trained not equal to 0 or n_layer")

    with tf.variable_scope('model/featurizer', reuse=reuse):
        bert = BertModel(
            config=bert_config,
            is_training=train,
            input_ids=input_ids,
            input_mask=mask,
            token_type_ids=token_type_ids,
            use_one_hot_embeddings=False,
            scope=None
        )
        output_state = {
            'embed_weights': bert.get_embedding_table(),
            'features': tf.reshape(
                bert.get_pooled_output(),
                shape=tf.concat((initial_shape[: -2], [config.n_embed]), 0)
            ),
            'sequence_features': tf.reshape(
                bert.get_sequence_output(),
                shape=tf.concat((initial_shape[:-1], [config.n_embed]), 0)
            ),
            'pool_idx': lengths
        }
        if config.num_layers_trained == 0:
            output_state = {k: tf.stop_gradient(v) for k, v in output_state.items()}

        return output_state



import functools

import numpy as np
import tensorflow as tf

from finetune.optimizers.recompute_grads import recompute_grad
from finetune.util.shapes import shape_list
from finetune.nn.activations import act_fns


def norm(x, scope, axis=[-1], e=1e-5):
    with tf.variable_scope(scope):
        n_state = shape_list(x)[-1]
        g = tf.get_variable("g", [n_state], initializer=tf.constant_initializer(1))
        b = tf.get_variable("b", [n_state], initializer=tf.constant_initializer(0))
        u = tf.reduce_mean(x, axis=axis, keepdims=True)
        s = tf.reduce_mean(tf.square(x - u), axis=axis, keepdims=True)
        x = (x - u) * tf.rsqrt(s + e)
        x = x * g + b
        return x


def dropout(x, pdrop, train):
    if train and pdrop > 0:
        x = tf.nn.dropout(x, 1 - pdrop)
    return x


def mask_attn_weights(w):
    n = shape_list(w)[-1]
    b = tf.matrix_band_part(tf.ones([n, n]), -1, 0)
    b = tf.reshape(b, [1, 1, n, n])
    w = w * b + -1e9 * (1 - b)
    return w


def attn_weights(q, k, v, attn_pdrop, train=False, scale=False, mask=True):
    w = tf.matmul(q, k)

    if scale:
        n_state = shape_list(v)[-1]
        w = w * tf.rsqrt(tf.cast(n_state, tf.float32))

    if mask:
        w = mask_attn_weights(w)
    w = tf.nn.softmax(w)
    return w


def split_states(x, n):
    x_shape = shape_list(x)
    m = x_shape[-1]
    new_x_shape = x_shape[:-1] + [n, m // n]
    return tf.reshape(x, new_x_shape)


def merge_states(x):
    x_shape = shape_list(x)
    new_x_shape = x_shape[:-2] + [np.prod(x_shape[-2:])]
    return tf.reshape(x, new_x_shape)


def split_heads(x, n, k=False):
    if k:
        return tf.transpose(split_states(x, n), [0, 2, 3, 1])
    else:
        return tf.transpose(split_states(x, n), [0, 2, 1, 3])


def merge_heads(x):
    return merge_states(tf.transpose(x, [0, 2, 1, 3]))


def conv1d(x, scope, nf, rf, w_init=tf.random_normal_initializer(stddev=0.02), b_init=tf.constant_initializer(0),
           pad='VALID', train=False):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        w = tf.get_variable("w", [rf, nx, nf], initializer=w_init)
        b = tf.get_variable("b", [nf], initializer=b_init)
        if rf == 1:  # faster 1x1 conv
            c = tf.reshape(tf.matmul(tf.reshape(x, [-1, nx]), tf.reshape(w, [-1, nf])) + b, shape_list(x)[:-1] + [nf])
        else:  # was used to train LM
            c = tf.nn.conv1d(x, w, stride=1, padding=pad) + b
        return c

def multihead_qkv(x, n_state, n_head, train):
    c = conv1d(x, 'c_attn', n_state * 3, 1, train=train)
    q, k, v = tf.split(c, 3, 2)
    q = split_heads(q, n_head)
    k = split_heads(k, n_head, k=True)
    v = split_heads(v, n_head)
    return q, k, v


def attn(x, scope, n_state, n_head, resid_pdrop, attn_pdrop, train=False, scale=False, mask=True):
    assert n_state % n_head == 0
    with tf.variable_scope(scope):
        q, k, v = multihead_qkv(x, n_state, n_head, train)
        w = attn_weights(q, k, v, attn_pdrop=attn_pdrop, train=train, scale=scale,
                        mask=mask)
        w = dropout(w, attn_pdrop, train)

        a = tf.matmul(w, v)

        a = merge_heads(a)
        a = conv1d(a, 'c_proj', n_state, 1, train=train)
        a = dropout(a, resid_pdrop, train)
        return a


def mlp(x, scope, n_state, act_fn, resid_pdrop, train=False):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        act = act_fns[act_fn]
        h = act(conv1d(x, 'c_fc', n_state, 1, train=train))
        h2 = conv1d(h, 'c_proj', nx, 1, train=train)
        h2 = dropout(h2, resid_pdrop, train)
        return h2


def block(x, n_head, act_fn, resid_pdrop, attn_pdrop, scope, train=False, scale=False):
    with tf.variable_scope(scope):
        nx = shape_list(x)[-1]
        a = attn(x, 'attn', nx, n_head, resid_pdrop, attn_pdrop, train=train, scale=scale)
        n = norm(x + a, 'ln_1')
        m = mlp(n, 'mlp', nx * 4, act_fn, resid_pdrop, train=train)
        h = norm(n + m, 'ln_2')
        return h


def embed(X, we):
    e = tf.gather(we, X)
    h = tf.reduce_sum(e, 2)
    return h


def gpt_featurizer(X, encoder, config, train=False, reuse=None):
    """
    The transformer element of the finetuning model. Maps from tokens ids to a dense, embedding of the sequence.

    :param X: A tensor of token indexes with shape [batch_size, sequence_length, token_idx]
    :param encoder: A TextEncoder object.
    :param config: A config object, containing all parameters for the featurizer.
    :param train: If this flag is true, dropout and losses are added to the graph.
    :param reuse: Should reuse be set within this scope.
    :return: A dict containing;
        embed_weights: the word embedding matrix.
        features: The output of the featurizer_final state.
        sequence_features: The output of the featurizer at each timestep.
    """
    initial_shape = tf.shape(X)
    X = tf.reshape(X, shape=tf.concat(([-1], initial_shape[-2:]), 0))

    with tf.variable_scope('model/featurizer', reuse=reuse):
        embed_weights = tf.get_variable(
            name="we",
            shape=[encoder.vocab_size + config.max_length, config.n_embed],
            initializer=tf.random_normal_initializer(stddev=config.weight_stddev)
        )
        if config.train_embeddings:
            embed_weights = dropout(embed_weights, config.embed_p_drop, train)
        else:
            embed_weights = tf.stop_gradient(embed_weights)

        X = tf.reshape(X, [-1, config.max_length, 2])

        h = embed(X, embed_weights)
        for layer in range(config.n_layer):
            if (config.n_layer - layer) == config.num_layers_trained and config.num_layers_trained != config.n_layer:
                h = tf.stop_gradient(h)
                train_layer = False
            else:
                train_layer = train

            with tf.variable_scope('h%d_' % layer):
                block_fn = functools.partial(block, n_head=config.n_heads, act_fn=config.act_fn,
                                             resid_pdrop=config.resid_p_drop, attn_pdrop=config.attn_p_drop,
                                             scope='h%d' % layer, train=train_layer, scale=True)
                if config.low_memory_mode and train_layer:
                    block_fn = recompute_grad(block_fn, use_entire_scope=True)
                if layer < config.n_layer - 1:
                    h = block_fn(h)
                else:
                    h_out = block_fn(h)

            # get the attention weights from the last layer
            if layer == config.n_layer - 1:
                with tf.variable_scope('h%d_/h%d/attn' % (layer, layer), reuse=True):
                    q, k, v = multihead_qkv(h, n_state=shape_list(h)[-1], n_head=config.n_heads, train=train)
                    w = attn_weights(q, k, v, attn_pdrop=config.attn_p_drop, train=train_layer, scale=True)


        # Use hidden state at classifier token as input to final proj. + softmax
        clf_h = tf.reshape(h_out, [-1, config.n_embed])  # [batch * seq_len, embed]
        clf_token = encoder['_classify_']
        pool_idx = tf.cast(tf.argmax(tf.cast(tf.equal(X[:, :, 0], clf_token), tf.float32), 1), tf.int32)
        clf_h = tf.gather(clf_h, tf.range(shape_list(X)[0], dtype=tf.int32) * config.max_length + pool_idx)
        clf_h = tf.reshape(clf_h, shape=tf.concat((initial_shape[: -2], [config.n_embed]), 0))
        seq_feats = tf.reshape(h, shape=tf.concat((initial_shape[:-1], [config.n_embed]), 0))

        return {
            'embed_weights': embed_weights,
            'features': clf_h,
            'sequence_features': seq_feats,
            'pool_idx': pool_idx,
            'attention_weights': w  # [n_heads, seq_len, seq_len]
        }



import numpy as np

from finetune.base import BaseModel
from finetune.input_pipeline import BasePipeline
from finetune.encoding.input_encoder import ArrayEncodedOutput
from finetune.encoding.target_encoders import IDEncoder
import tensorflow as tf

from finetune.nn.target_blocks import multi_choice_question
from finetune.util import list_transpose

class MultipleChoicePipeline(BasePipeline):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _post_data_initialization(self, Y):
        super()._post_data_initialization(Y)
        self.target_dim = self.target_dim_

    def _text_to_ids(self, Xs, Y=None, pad_token=None):
        """
        Format multi question examples as a list of IDs
        """
        q, answer_list = Xs

        pairs = [[q, answer_list[idx]] for idx in range(len(answer_list))]
        arrays = []
        for pair in pairs:
            arrays.append(next(super()._text_to_ids(pair, Y=Y)))

        kwargs = arrays[0]._asdict()
        kwargs['tokens'] = [arr.tokens for arr in arrays]
        kwargs['token_ids'] = np.stack([arr.token_ids for arr in arrays], 0)
        kwargs['mask'] = np.stack([arr.mask for arr in arrays], 0)
        yield ArrayEncodedOutput(**kwargs)

    def _format_for_encoding(self, X):
        return [X]

    def feed_shape_type_def(self):
        TS = tf.TensorShape
        return ({"tokens": tf.int32, "mask": tf.float32}, tf.int32), (
            {"tokens": TS([self.target_dim, self.config.max_length, 2]), "mask": TS([self.target_dim, self.config.max_length])}, TS([]))

    def _target_encoder(self):
        return IDEncoder()


class MultipleChoice(BaseModel):
    """
    Multi choice question finetune model.

    :param config: A :py:class:`finetune.config.Settings` object or None (for default config).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_answers = None

    def _get_input_pipeline(self):
        return MultipleChoicePipeline(self.config)

    def finetune(self, questions, answers, correct_answer, fit_lm_only=False):
        """
        :param questions: List or array of text, shape [batch]
        :param answers: List or array of text, shape [batch, n_answers], must contain the correct answer for each entry.
        :param correct_answer: List or array of correct answers [batch] either in the format of an idx to the correct
                answer or a string of the correct answer.
        :param batch_size: integer number of examples per batch. When N_GPUS > 1, this number
                           corresponds to the number of training examples provided to each GPU.
        """
        answer_idx = []
        if not len(correct_answer) == len(answers) == len(questions):
            raise ValueError("Answers, questions and corrext_answer are not all the same length, {},{},{}".format(
                len(questions), len(correct_answer), len(answers)
            ))

        for correct, others in zip(correct_answer, answers):
            if isinstance(correct, int):
                if 0 > correct > len(others):
                    raise ValueError(
                        "Correct answer is of type int but is invalid with value {} for answers of len {}".format(
                            correct, len(others)))
                answer_idx.append(correct)
            else:
                try:
                    ans_idx = others.index(correct)
                    answer_idx.append(ans_idx)
                except ValueError:
                    raise ValueError(
                        "Correct answer {} is not contained in possible answers {}".format(correct, others))

        labels = None if fit_lm_only else answer_idx
        self.input_pipeline.target_dim_ = len(answers[0])
        return super().finetune(list(zip(questions, answers)), Y=labels)

    @staticmethod
    def _target_model(config, featurizer_state, targets, n_outputs, train=False, reuse=None, **kwargs):
        return multi_choice_question(
            hidden=featurizer_state['features'],
            targets=targets,
            n_targets=n_outputs,
            config=config,
            train=train,
            reuse=reuse,
            **kwargs
        )

    def _predict_op(self, logits, **kwargs):
        return tf.argmax(logits, -1)

    def _predict_proba_op(self, logits, **kwargs):
        return tf.nn.softmax(logits, -1)

    def predict(self, questions, answers):
        """
        Produces a list of most likely class labels as determined by the fine-tuned model.


        :param question: List or array of text, shape [batch]
        :param answers: List or array of text, shape [batch, n_answers]
        :returns: list of class labels.
        """
        raw_ids = BaseModel.predict(self, list(zip(questions, answers)))
        return [ans[i] for ans, i in zip(answers, raw_ids)]

    def predict_proba(self, questions, answers):
        """
        Produces a probability distribution over classes for each example in X.


        :param question: List or array of text, shape [batch]
        :param answers: List or array of text, shape [batch, n_answers]
        :returns: list of dictionaries.  Each dictionary maps from a class label to its assigned class probability.
        """
        answers = list_transpose(answers)
        raw_probas = self._predict_proba(zip(questions, answers))

        formatted_predictions = []
        for probas, *answers_per_sample in zip(raw_probas, *answers):
            formatted_predictions.append(
                dict(zip(answers_per_sample, probas))
            )
        return formatted_predictions

    def featurize(self, questions, answers):
        """
        Embeds inputs in learned feature space. Can be called before or after calling :meth:`finetune`.

        :param questions: List or array of text, shape [batch]
        :param answers: List or array of text, shape [n_answers, batch]
        :returns: np.array of features of shape (n_examples, embedding_size).
        """
        return BaseModel.featurize(self, zip(questions, answers))



import itertools
import logging
import sys
import math
import os

from abc import ABCMeta, abstractmethod

import tqdm
import numpy as np
import tensorflow as tf
from tensorflow.python.data import Dataset
from sklearn.model_selection import train_test_split


import finetune
from finetune.errors import FinetuneError
from finetune.encoding.input_encoder import ArrayEncodedOutput, EncodedOutput
from finetune.util.imbalance import compute_class_weights

LOGGER = logging.getLogger('finetune')


class BasePipeline(metaclass=ABCMeta):
    def __init__(self, config):
        self.config = config
        self.text_encoder = self.config.base_model.get_encoder()

        self.label_encoder = None
        self.target_dim = None
        self.pad_idx_ = None
        self.rebuild = False
        self.epoch = 0

    @property
    def dataset_size(self):
        return self.config.dataset_size

    @abstractmethod
    def _target_encoder(self):
        # Overridden by subclass to produce the right target encoding for a given target model.
        raise NotImplementedError

    def feed_shape_type_def(self):
        TS = tf.TensorShape
        return (
            (
                {
                    "tokens": tf.int32,
                    "mask": tf.float32
                },
                tf.float32
            ),
            (
                {
                    "tokens": TS([self.config.max_length, 2]),
                    "mask": TS([self.config.max_length])
                },
                TS([self.target_dim])
            )
        )

    def _array_format(self, encoded_output, pad_token=None):
        """
        Returns numpy array of token idxs and corresponding mask
        Returned `x` array contains two channels:
            0: byte-pair encoding embedding
            1: positional embedding
        """
        seq_length = len(encoded_output.token_ids)
        x = np.zeros((self.config.max_length, 2), dtype=np.int32)
        mask = np.zeros((self.config.max_length), dtype=np.float32)

        if encoded_output.labels is not None:
            labels_arr = np.empty((self.config.max_length), dtype='object')
            labels_arr.fill((pad_token or self.config.pad_token))
        else:
            labels_arr = None

        # BPE embedding
        x[:seq_length, 0] = encoded_output.token_ids
        # masking: value of 1 means "consider this in cross-entropy LM loss"
        mask[1:seq_length] = 1
        if encoded_output.labels:
            labels_arr[:seq_length] = encoded_output.labels
        # positional_embeddings
        x[:, 1] = np.arange(
            self.text_encoder.vocab_size, self.text_encoder.vocab_size + self.config.max_length
        )

        output = ArrayEncodedOutput(
            token_ids=x,
            tokens=encoded_output.tokens,
            labels=labels_arr,
            char_locs=encoded_output.char_locs,
            mask=mask,
        )
        return output

    def text_to_tokens_mask(self, X, Y=None):
        out_gen = self._text_to_ids(X, pad_token=self.config.pad_token)
        for out in out_gen:
            feats = {"tokens": out.token_ids, "mask": out.mask}
            if Y is None:
                yield feats
            else:
                yield feats, self.label_encoder.transform([Y])[0]

    def _post_data_initialization(self, Y):
        self.label_encoder = self._target_encoder()
        if not callable(Y):
            Y_fit = Y
            self.label_encoder.fit(Y)
        else:
            Y_fit = list(itertools.islice(Y(), 10000))
            self.label_encoder.fit(Y_fit)
        self.config.pad_idx = self.pad_idx

        target_dim = self.label_encoder.target_dim
        self.lm_loss_coef = self.config.lm_loss_coef if target_dim is not None else 1.0
        self.target_dim = target_dim

        if Y_fit is not None:
            self.config.class_weights = compute_class_weights(class_weights=self.config.class_weights, Y=Y_fit)

    def _dataset_with_targets(self, Xs, Y, train):
        if not callable(Xs) and not callable(Y):
            dataset = lambda: zip(Xs, Y)
        elif callable(Xs) and callable(Y):
            dataset = lambda: zip(Xs(), Y())  # encode one sample at a time.
        else:
            raise ValueError("Either neither or both of Xs and Y should be callable, not a mixture")

        dataset_encoded = lambda: itertools.chain.from_iterable(
            map(lambda xy: self.text_to_tokens_mask(*xy), dataset()))
        shape_def = self.feed_shape_type_def()
        if not callable(Y) and self.config.chunk_long_sequences:
            dataset_encoded_list = list(dataset_encoded())  # come up with a more principled way to do this .
            self.config.dataset_size = len(dataset_encoded_list)
        return Dataset.from_generator(lambda: self.wrap_tqdm(dataset_encoded(), train), *shape_def)

    def _dataset_without_targets(self, Xs, train):
        if not callable(Xs):
            Xs_fn = lambda: self.wrap_tqdm(Xs, train)
        else:
            Xs_fn = lambda: self.wrap_tqdm(Xs(), train)

        dataset_encoded = lambda: itertools.chain.from_iterable(map(self.text_to_tokens_mask, Xs_fn()))
        types, shapes = self.feed_shape_type_def()
        return Dataset.from_generator(dataset_encoded, types[0], shapes[0])  # 0s cut out the targets

    def _integer_val_size(self, val_size):
        if isinstance(val_size, float):
            return int(val_size * self.config.dataset_size)
        return val_size

    def validation_settings(self, n_examples, batch_size):
        """
        Auto-select reasonable validation settings
        """
        if self.config.val_size is not None and self.config.val_interval is not None:
            return self._integer_val_size(self.config.val_size), self.config.val_interval

        # Auto-select reasonable validation size
        if self.config.val_size is None:
            if n_examples < 50:
                val_size = 0
            else:
                val_size = max(5, int(0.05 * n_examples))
                val_size = min(100, val_size)
        else:
            val_size = self._integer_val_size(self.config.val_size)

        # Auto-select reasonable validation interval
        if self.config.val_interval is None:
            # sys.maxsize corresponds to never running validation
            # and is used when val_size is set to 0
            val_interval = 4 * int(math.ceil(val_size / batch_size)) or sys.maxsize
        else:
            val_interval = self.config.val_interval

        return int(val_size), int(val_interval)

    def resampling(self, Xs, Y):
        return Xs, Y

    def _make_dataset(self, Xs, Y, train=False):
        if Y is not None:
            dataset = lambda: self._dataset_with_targets(Xs, Y, train=train)
        else:
            dataset = lambda: self._dataset_without_targets(Xs, train=train)
        return dataset

    def wrap_tqdm(self, gen, train):
        if self.config.debugging_logs:
            return gen

        if train is None:
            return gen

        try:
            total = len(gen)
        except:
            if train:
                total = self.config.dataset_size
            else:
                total = self.config.val_size

        def internal_gen():
            current_epoch = (self.epoch - 1) % self.config.n_epochs + 1
            it = iter(gen)

            if train:
                if self.config.prefit_init and self.epoch <= self.config.n_epochs:
                    desc = "Initialization Epoch {}/{}".format(current_epoch, self.config.n_epochs)
                else:
                    desc = "Epoch {}/{}".format(current_epoch, self.config.n_epochs)
            else:
                desc = "Validation"
            for _, i in zip(range(self._skip_tqdm), it):
                yield i

            for i in tqdm.tqdm(it, desc=desc, total=total, miniters=1, leave=current_epoch == self.config.n_epochs and train):
                yield i

            if train:
                self.epoch += 1

        return internal_gen()

    def get_train_input_fns(self, Xs, Y=None, batch_size=None, val_size=None):
        self.epoch = 1
        batch_size = batch_size or self.config.batch_size

        shuffle_buffer_size = self.config.shuffle_buffer_size
        val_size = val_size or 0
        prefetch_buffer = 2  # breaks the pipeline to allow concurrency

        if callable(Xs):
            try:
                self.config.dataset_size = len(Xs())
            except TypeError:
                if self.config.dataset_size is None:
                    raise FinetuneError(
                        "Generator input function does not have a length and no `config.dataset_size` is specified. "
                        "You must set `config.dataset_size` explicitly."
                    )
        else:
            self.config.dataset_size = len(Xs)

        self.config.val_size, self.config.val_interval = self.validation_settings(
            n_examples=len(Xs) if not callable(Xs) else self.config.dataset_size,
            batch_size=batch_size or self.config.batch_size
        )
        self.config.dataset_size -= val_size

        if Y is not None:
            self._post_data_initialization(Y)

        if callable(Xs) or Y is None:
            self._skip_tqdm = val_size
            dataset = self._make_dataset(Xs, Y, train=True)
            val_dataset_unbatched = lambda: dataset().shuffle(
                shuffle_buffer_size, seed=self.config.seed, reshuffle_each_iteration=False
            ).take(self.config.val_size)
            train_dataset_unbatched = lambda: dataset().shuffle(
                shuffle_buffer_size, seed=self.config.seed, reshuffle_each_iteration=False
            ).skip(self.config.val_size)
        else:
            self._skip_tqdm = 0
            if self.config.val_set is None:
                Xs_tr, Xs_va, Y_tr, Y_va = train_test_split(Xs, Y, test_size=self.config.val_size, random_state=self.config.seed)
            else:
                Xs_tr, Y_tr = Xs, Y
                Xs_va, Y_va = self.config.val_set

            Xs_tr, Y_tr = self.resampling(Xs_tr, Y_tr)
            self.config.dataset_size = len(Xs_tr)
            val_dataset_unbatched = self._make_dataset(Xs_va, Y_va, train=False)
            train_dataset_unbatched = self._make_dataset(Xs_tr, Y_tr, train=True)

        if self.config.chunk_long_sequences:
            train_dataset_unbatched()

        val_dataset = lambda: val_dataset_unbatched().batch(batch_size, drop_remainder=False).cache().prefetch(prefetch_buffer)
        train_dataset = lambda: train_dataset_unbatched().batch(batch_size, drop_remainder=False).repeat(
            self.config.n_epochs).prefetch(prefetch_buffer)

        return val_dataset, train_dataset, self.config.val_size, self.config.val_interval

    def get_predict_input_fn(self, Xs, batch_size=None):
        batch_size = batch_size or self.config.batch_size
        tf_dataset = lambda: self._dataset_without_targets(Xs, train=None).batch(batch_size)
        return tf_dataset

    @property
    def pad_idx(self):
        if self.pad_idx_ is None:
            if hasattr(self.label_encoder, "classes_"):
                classes = list(self.label_encoder.classes_)
                if self.config.pad_token in classes:
                    self.pad_idx_ = classes.index(self.config.pad_token)
                else:
                    self.pad_idx_ = None
        return self.pad_idx_

    def _format_for_encoding(self, X):
        """
        Most subclasses take in inputs as:
            List (batch) of list (docs)

        Encode_multi_input expect the following format:
            List (batch) of list (docs) of list (subseqs) of text

        This method is responsible for standardizing inputs to the above format
        """
        return [[X]]

    def _format_for_inference(self, X):
        return list(X)

    def _text_to_ids(self, Xs, Y=None, pad_token=None):
        Xs = self._format_for_encoding(Xs)
        if self.config.chunk_long_sequences and len(Xs) == 1:
            # can only chunk single sequence inputs
            chunk_size = self.config.max_length - 2
            step_size = chunk_size // 3
            encoded = self.text_encoder.encode_multi_input(
                Xs,
                Y=Y,
                max_length=sys.maxsize,
                pad_token=(pad_token or self.config.pad_token)
            )
            length = len(encoded.token_ids)
            starts = list(range(0, length, step_size))
            for start in starts:
                d = dict()
                end = start + chunk_size
                for field in EncodedOutput._fields:
                    field_value = getattr(encoded, field)
                    if field_value is not None:
                        d[field] = field_value[start:end]
                yield self._array_format(EncodedOutput(**d), pad_token=pad_token)
        else:
            encoder_out = self.text_encoder.encode_multi_input(
                Xs,
                Y=Y,
                max_length=self.config.max_length,
                pad_token=(pad_token or self.config.pad_token)
            )

            yield self._array_format(encoder_out, pad_token=(pad_token or self.config.pad_token))



import numpy as np
import tensorflow as tf


def np_softmax(x, t=1, axis=-1):
    x = x / t
    x = x - np.max(x, axis=axis, keepdims=True)
    ex = np.exp(x)
    return ex / np.sum(ex, axis=axis, keepdims=True)


def viterbi_decode(score, transition_params):
    """Decode the highest scoring sequence of tags outside of TensorFlow.
    This should only be used at test time.
    Args:
        score: A [seq_len, num_tags] matrix of unary potentials.
        transition_params: A [num_tags, num_tags] matrix of binary potentials.
    Returns:
        viterbi: A [seq_len] list of integers containing the highest scoring tag
            indices.
        viterbi_score: A float containing the score for the Viterbi sequence.
    """
    trellis = np.zeros_like(score)
    backpointers = np.zeros_like(score, dtype=np.int32)
    trellis[0] = score[0]

    for t in range(1, score.shape[0]):
        v = np.expand_dims(trellis[t - 1], 1) + transition_params
        trellis[t] = score[t] + np.max(v, 0)
        backpointers[t] = np.argmax(v, 0)

    viterbi = [np.argmax(trellis[-1])]
    for bp in reversed(backpointers[1:]):
        viterbi.append(bp[viterbi[-1]])
    viterbi.reverse()

    return viterbi, np_softmax(trellis, axis=-1)


def sequence_decode(logits, transition_matrix):
    """ A simple py_func wrapper around the Viterbi decode allowing it to be included in the tensorflow graph. """

    def _sequence_decode(logits, transition_matrix):
        all_predictions = []
        all_logits = []
        for logit in logits:
            viterbi_sequence, viterbi_logits = viterbi_decode(logit, transition_matrix)
            all_predictions.append(viterbi_sequence)
            all_logits.append(viterbi_logits)
        return np.array(all_predictions, dtype=np.int32), np.array(all_logits, dtype=np.float32)

    return tf.py_func(_sequence_decode, [logits, transition_matrix], [tf.int32, tf.float32])



"""
Modified from https://github.com/openai/iaf/blob/master/tf_utils/adamax.py
"""
__license__ = """The MIT License (MIT)

Copyright (c) 2016 openai

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import tensorflow as tf


class AdamaxOptimizer(optimizer.Optimizer):
    """
    Optimizer that implements the Adamax algorithm.

    See [Kingma et. al., 2014](http://arxiv.org/abs/1412.6980)
    ([pdf](http://arxiv.org/pdf/1412.6980.pdf)).
    """

    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, use_locking=False, name="Adamax", epsilon=1e-8):
        super(AdamaxOptimizer, self).__init__(use_locking, name)
        self._lr = learning_rate
        self._beta1 = beta1
        self._beta2 = beta2
        self.eps = epsilon

        # Tensor versions of the constructor arguments, created in _prepare().
        self._lr_t = None
        self._beta1_t = None
        self._beta2_t = None

    def _prepare(self):
        self._lr_t = ops.convert_to_tensor(self._lr, name="learning_rate")
        self._beta1_t = ops.convert_to_tensor(self._beta1, name="beta1")
        self._beta2_t = ops.convert_to_tensor(self._beta2, name="beta2")

    def _create_slots(self, var_list):
        # Create slots for the first and second moments.
        for v in var_list:
            self._zeros_slot(v, "m", self._name)
            self._zeros_slot(v, "v", self._name)

    def _apply_dense(self, grad, var):
        lr_t = math_ops.cast(self._lr_t, var.dtype.base_dtype)
        beta1_t = math_ops.cast(self._beta1_t, var.dtype.base_dtype)
        beta2_t = math_ops.cast(self._beta2_t, var.dtype.base_dtype)

        v = self.get_slot(var, "v")
        v_t = v.assign(beta1_t * v + (1. - beta1_t) * grad)
        m = self.get_slot(var, "m")
        m_t = m.assign(tf.maximum(beta2_t * m + self.eps, tf.abs(grad)))
        g_t = v_t / m_t

        var_update = state_ops.assign_sub(var, lr_t * g_t)
        return control_flow_ops.group(*[var_update, m_t, v_t])

    def _resource_apply_dense(self, grad, var):
        return self._apply_dense(grad, var)

    def _resource_apply_sparse(self, grad, var):
        return self._apply_sparse(grad, var)

    def _apply_sparse(self, grad, var):
        return self.apply_dense(tf.convert_to_tensor(grad), var)


AdamaxWOptimizer = tf.contrib.opt.extend_with_decoupled_weight_decay(AdamaxOptimizer)